# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Collect reads containing a variant and split them into prefix, variant, and
suffix portions
"""

from __future__ import print_function, division, absolute_import
from collections import namedtuple
import logging

from .read_at_locus import read_at_locus_generator
from .default_parameters import (
    MIN_READ_MAPPING_QUALITY,
    USE_SECONDARY_ALIGNMENTS,
    USE_DUPLICATE_READS,
)
from .variant_helpers import trim_variant
from .dataframe_builder import DataFrameBuilder


VariantRead = namedtuple(
    "VariantRead", "prefix alt suffix name")


def trim_N_nucleotides(prefix, suffix):
    """
    Drop all occurrences of 'N' from prefix and suffix nucleotide strings
    by trimming.
    """
    if 'N' in prefix:
        # trim prefix to exclude all occurrences of N
        rightmost_index = prefix.rfind('N')
        logging.debug(
            "Trimming %d nucleotides from read prefix '%s'" % (
                rightmost_index + 1, prefix))
        prefix = prefix[rightmost_index + 1:]

    if 'N' in suffix:
        leftmost_index = suffix.find('N')
        logging.debug(
            "Trimming %d nucleotides from read suffix '%s'" % (
                len(suffix) - leftmost_index,
                suffix))
        suffix = suffix[:leftmost_index]

    return prefix, suffix

def convert_from_bytes_if_necessary(prefix, suffix):
    """
    Depending on how we extract data from pysam we may end up with either
    a string or a byte array of nucleotides. For consistency and simplicity,
    we want to only use strings in the rest of our code.
    """
    if isinstance(prefix, bytes):
        prefix = prefix.decode('ascii')

    if isinstance(suffix, bytes):
        suffix = suffix.decode('ascii')

    return prefix, suffix

def variant_read_from_single_read_at_locus(read, ref, alt):
    """
    Given a single ReadAtLocus object, return either a VariantRead or None
    (if the read's sequence didn't contain the variant nucleotides).

    Parameters
    ----------
    read : ReadAtLocus
        Read which may possibly contain the alternate nucleotides

    ref : str
        Reference sequence of the variant (empty for insertions)

    alt : str
        Alternate sequence of the variant (empty for deletions)

    """
    sequence = read.sequence
    reference_positions = read.reference_positions

    # positions of the nucleotides before and after the variant within
    # the read sequence
    read_pos_before = read.base0_read_position_before_variant
    read_pos_after = read.base0_read_position_after_variant

    # positions of the nucleotides before and after the variant on the
    # reference genome
    ref_pos_before = reference_positions[read_pos_before]

    if ref_pos_before is None:
        logging.warn(
            "Missing reference pos for nucleotide before variant on read: %s" % (
                read,))
        return None

    ref_pos_after = reference_positions[read_pos_after]

    if ref_pos_after is None:
        logging.warn(
            "Missing reference pos for nucleotide after variant on read: %s" % (
                read,))
        return None

    if len(ref) == 0:
        if ref_pos_after - ref_pos_before != 1:
            # if the number of nucleotides skipped isn't the same
            # as the number of reference nucleotides in the variant then
            # don't use this read
            logging.debug(
                "Positions before (%d) and after (%d) variant should be adjacent on read %s" % (
                    ref_pos_before,
                    ref_pos_after,
                    read))
            return None

        # insertions require a sequence of non-aligned bases
        # followed by the subsequence reference position
        ref_positions_for_inserted = reference_positions[
            read_pos_before + 1:read_pos_after]
        if any(insert_pos is not None for insert_pos in ref_positions_for_inserted):
            # all these inserted nucleotides should *not* align to the
            # reference
            logging.debug(
                "Skipping read, inserted nucleotides shouldn't map to reference")
            return None
    else:
        # substitutions and deletions
        if ref_pos_after - ref_pos_before != len(ref) + 1:
            # if the number of nucleotides skipped isn't the same
            # as the number of reference nucleotides in the variant then
            # don't use this read
            logging.debug(
                "Positions before (%d) and after (%d) variant should be adjacent on read %s" % (
                    ref_pos_before,
                    ref_pos_after,
                    read))
            return None

    nucleotides_at_variant_locus = sequence[read_pos_before + 1:read_pos_after]
    if nucleotides_at_variant_locus != alt:
        # read sequence doesn't match variant
        logging.debug("Read '%s' has '%s' at variant locus (%d:%d), required '%s'" % (
            read.name,
            nucleotides_at_variant_locus,
            read_pos_before + 1,
            read_pos_after,
            alt))
        return None

    prefix = sequence[:read_pos_before + 1]
    suffix = sequence[read_pos_after:]

    prefix, suffix = convert_from_bytes_if_necessary(prefix, suffix)
    prefix, suffix = trim_N_nucleotides(prefix, suffix)

    return VariantRead(
        prefix,
        nucleotides_at_variant_locus,
        suffix,
        name=read.name)

def variant_reads_from_reads_at_locus(reads, ref, alt):
    """
    Given a collection of ReadAtLocus objects, returns a
    list of VariantRead objects (which are split into prefix/variant/suffix
    nucleotides).

    Parameters
    ----------
    reads : sequence of ReadAtLocus records

    ref : str
        Reference sequence of the variant (empty for insertions)

    alt : str
        Alternate sequence of the variant (empty for deletions)

    Returns a list of VariantRead objects.
    """
    variant_reads = []
    for read in reads:
        variant_read = variant_read_from_single_read_at_locus(read, ref, alt)
        if variant_read is not None:
            variant_reads.append(variant_read)
    return variant_reads

def gather_reads_for_single_variant(
        samfile,
        variant,
        chromosome=None,
        use_duplicate_reads=USE_DUPLICATE_READS,
        use_secondary_alignments=USE_SECONDARY_ALIGNMENTS,
        min_mapping_quality=MIN_READ_MAPPING_QUALITY):
    """
    Find reads in the given SAM/BAM file which overlap the given variant, filter
    to only include those which agree with the variant's nucleotide(s), and turn
    them into a list of VariantRead objects.

    Parameters
    ----------
    samfile : pysam.AlignmentFile

    variant : varcode.Variant

    chromosome : str

    use_duplicate_reads : bool
        Should we use reads that have been marked as PCR duplicates

    use_secondary_alignments : bool
        Should we use reads at locations other than their best alignment

    min_mapping_quality : int
        Drop reads below this mapping quality

    Returns list of VariantRead objects.
    """
    logging.info("Gathering reads for %s" % variant)
    if chromosome is None:
        chromosome = variant.contig

    logging.info("Gathering variant reads for variant %s (chromosome=%s)" % (
        variant,
        chromosome))

    base1_position, ref, alt = trim_variant(variant)
    if len(ref) == 0:
        # if the variant is an insertion
        base1_position_before_variant = base1_position
        base1_position_after_variant = base1_position + 1
    else:
        base1_position_before_variant = base1_position - 1
        base1_position_after_variant = base1_position + len(ref)

    reads = list(read_at_locus_generator(
        samfile=samfile,
        chromosome=chromosome,
        base1_position_before_variant=base1_position_before_variant,
        base1_position_after_variant=base1_position_after_variant,
        use_duplicate_reads=use_duplicate_reads,
        use_secondary_alignments=use_secondary_alignments,
        min_mapping_quality=min_mapping_quality))

    logging.info(
        "Found %d reads at locus overlapping %s" % (len(reads), variant))

    variant_reads = list(
        variant_reads_from_reads_at_locus(
            reads=reads,
            ref=ref,
            alt=alt))
    logging.info("Found %d VariantReads for variant %s" % (
        len(variant_reads),
        variant))
    return variant_reads

def variant_reads_generator(
        variants,
        samfile,
        use_duplicate_reads=USE_DUPLICATE_READS,
        use_secondary_alignments=USE_SECONDARY_ALIGNMENTS,
        min_mapping_quality=MIN_READ_MAPPING_QUALITY):
    """
    Generates sequence of tuples, each containing a variant paired with
    a list of VariantRead objects.

    Parameters
    ----------
    variants : varcode.VariantCollection

    samfile : pysam.AlignmentFile

    use_duplicate_reads : bool
        Should we use reads that have been marked as PCR duplicates

    use_secondary_alignments : bool
        Should we use reads at locations other than their best alignment

    min_mapping_quality : int
        Drop reads below this mapping quality
    """
    chromosome_names = set(samfile.references)
    for variant in variants:
        # I imagine the conversation went like this:
        # A: "Hey, I have an awesome idea"
        # B: "What's up?"
        # A: "Let's make two nearly identical reference genomes"
        # B: "But...that sounds like it might confuse people."
        # A: "Nah, it's cool, we'll give the chromosomes different prefixes!"
        # B: "OK, sounds like a good idea."
        if variant.contig in chromosome_names:
            chromosome = variant.contig
        elif "chr" + variant.contig in chromosome_names:
            chromosome = "chr" + variant.contig
        else:
            logging.warn(
                "Chromosome '%s' from variant %s not in alignment file %s" % (
                    chromosome, variant, samfile.filename))
            continue
        variant_reads = gather_reads_for_single_variant(
            samfile=samfile,
            chromosome=chromosome,
            variant=variant,
            use_duplicate_reads=use_duplicate_reads,
            use_secondary_alignments=use_secondary_alignments,
            min_mapping_quality=min_mapping_quality)
        yield variant, variant_reads


def variant_reads_dataframe(
        variants,
        samfile,
        use_duplicate_reads=USE_DUPLICATE_READS,
        use_secondary_alignments=USE_SECONDARY_ALIGNMENTS,
        min_mapping_quality=MIN_READ_MAPPING_QUALITY):
    """
    Creates a DataFrame containing sequences around variant from variant
    collection and BAM/SAM file.

    Parameters
    ----------
    variants : varcode.VariantCollection

    samfile : pysam.AlignmentFile

    use_duplicate_reads : bool
        Should we use reads that have been marked as PCR duplicates

    use_secondary_alignments : bool
        Should we use reads at locations other than their best alignment

    min_mapping_quality : int
        Drop reads below this mapping quality
    """
    df_builder = DataFrameBuilder(VariantRead)
    for variant, variant_reads in variant_reads_generator(
            variants,
            samfile,
            use_duplicate_reads=use_duplicate_reads,
            use_secondary_alignments=use_secondary_alignments,
            min_mapping_quality=min_mapping_quality):
        for variant_read in variant_reads:
            df_builder.add(variant, variant_read)
    return df_builder.to_dataframe()



#!/usr/bin/env python

# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Translate each non-synonymous coding variants into possible mutant protein
sequences using an RNAseq BAM from the same tissuie.
"""

from __future__ import print_function, division, absolute_import

import argparse

import varcode
from pysam import AlignmentFile

from isovar.args import add_somatic_vcf_args, add_rna_args
from isovar.translation import translate_variants_dataframe
from isovar.default_parameters import (
    MIN_TRANSCRIPT_PREFIX_LENGTH,
    MAX_REFERENCE_TRANSCRIPT_MISMATCHES,
    PROTEIN_SEQUENCE_LENGTH,
)


parser = argparse.ArgumentParser()
parser = add_somatic_vcf_args(parser)
parser = add_rna_args(parser)

parser.add_argument(
    "--protein-sequence-length",
    default=PROTEIN_SEQUENCE_LENGTH,
    type=int)

parser.add_argument(
    "--max-reference-transcript-mismatches",
    type=int,
    default=MAX_REFERENCE_TRANSCRIPT_MISMATCHES)

parser.add_argument(
    "--min-transcript-prefix-length",
    type=int,
    default=MIN_TRANSCRIPT_PREFIX_LENGTH,
    help=(
        "Number of nucleotides before the variant we try to match against "
        "a reference transcript. Values greater than zero exclude variants "
        "near the start codon of transcripts without 5' UTRs."))

parser.add_argument(
    "--output",
    default="isovar-translate-variants-results.csv",
    help="Name of CSV file which contains predicted sequences")

if __name__ == "__main__":
    args = parser.parse_args()

    print(args)

    variants = varcode.load_vcf(
        args.vcf,
        genome=args.genome)

    samfile = AlignmentFile(args.bam)
    df = translate_variants_dataframe(
        variants=variants,
        samfile=samfile,
        protein_sequence_length=args.protein_sequence_length,
        min_reads_supporting_rna_sequence=args.min_reads,
        min_transcript_prefix_length=args.min_transcript_prefix_length,
        max_transcript_mismatches=args.max_reference_transcript_mismatches)
    print(df)
    df.to_csv(args.output)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from isovar.variant_read import gather_reads_for_single_variant
from pyensembl import ensembl_grch38

from varcode import Variant
from nose.tools import eq_

from testing_helpers import load_bam

def test_partition_variant_reads_snv():
    samfile = load_bam("data/cancer-wgs-primary.chr12.bam")
    chromosome = "chr12"
    base1_location = 65857041
    ref = "G"
    alt = "C"
    variant = Variant(
        contig=chromosome,
        start=base1_location,
        ref=ref,
        alt=alt,
        ensembl=ensembl_grch38)
    variant_reads = gather_reads_for_single_variant(
        samfile=samfile,
        chromosome=chromosome,
        variant=variant)
    assert len(variant_reads) > 1
    for variant_read in variant_reads:
        eq_(variant_read.alt, alt)

def test_partition_variant_reads_deletion():
    samfile = load_bam("data/cancer-wgs-primary.chr12.bam")
    chromosome = "chr12"
    base1_location = 70091490
    ref = "TTGTAGATGCTGCCTCTCC"
    alt = ""
    variant = Variant(
        contig=chromosome,
        start=base1_location,
        ref=ref,
        alt=alt,
        ensembl=ensembl_grch38)
    variant_reads = gather_reads_for_single_variant(
        samfile=samfile,
        chromosome=chromosome,
        variant=variant)
    assert len(variant_reads) > 1
    for variant_read in variant_reads:
        eq_(variant_read.alt, alt)

if __name__ == "__main__":
    test_partition_variant_reads_snv()
    test_partition_variant_reads_deletion()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import
from collections import OrderedDict
from six import integer_types, text_type, binary_type

from varcode import Variant
import pandas as pd

VALID_ELEMENT_TYPES = integer_types + (text_type, binary_type, float, bool)

class DataFrameBuilder(object):
    """
    Helper class for constructing a DataFrame which always has fields
    of a variant (chr/pos/ref/alt) as well as some subset of the fields
    from a namedtuple.
    """
    def __init__(
            self,
            element_class,
            exclude=set([]),
            converters={},
            rename_dict={},
            variant_columns=True):
        """
        Parameters
        ----------
        element_class : type
            Expected to a have a class-member named '_fields' which is a list
            of field names.

        exclude : set
            Field names from element_class which should be used as columns for
            the DataFrame we're building

        converters : dict
            Dictionary of names mapping to functions. These functions will be
            applied to each element of a column before it's added to the
            DataFrame.

        rename_dict : dict
            Dictionary mapping element_class field names to desired column names
            in the produced DataFrame.

        variant_columns : bool
            If True, then add four columns for fields of a Variant: chr/pos/ref/alt
        """
        self.element_class = element_class
        self.rename_dict = rename_dict
        self.converters = converters
        self.variant_columns = variant_columns

        # remove specified field names without changing the order of the others
        self.original_field_names = [
            x
            for x in element_class._fields
            if x not in exclude
        ]

        for name in converters:
            if name not in self.original_field_names:
                raise ValueError("No field named '%s', valid names: %s" % (
                    name,
                    self.original_field_names))

        self.renamed_field_names = [
            self.rename_dict.get(x, x)
            for x in self.original_field_names
        ]
        if self.variant_columns:
            columns_list = [
                # fields related to variant
                ("chr", []),
                ("pos", []),
                ("ref", []),
                ("alt", []),
            ]
        else:
            columns_list = []

        for name in self.renamed_field_names:
            columns_list.append((name, []))

        self.columns_dict = OrderedDict(columns_list)

    def add(self, variant, element):
        if self.variant_columns:
            assert isinstance(variant, Variant)
            self.columns_dict["chr"].append(variant.contig)
            self.columns_dict["pos"].append(variant.original_start)
            self.columns_dict["ref"].append(variant.original_ref)
            self.columns_dict["alt"].append(variant.original_alt)
        else:
            assert variant is None

        assert isinstance(element, self.element_class)

        for name in self.original_field_names:
            value = getattr(element, name)

            if name in self.converters:
                fn = self.converters[name]
                value = fn(value)

            if not isinstance(value, VALID_ELEMENT_TYPES):
                raise ValueError(
                    "Please provider converter for field '%s' : %s to make a scalar or string" % (
                        name,
                        type(value)))

            if name in self.rename_dict:
                name = self.rename_dict[name]
            self.columns_dict[name].append(value)

    def to_dataframe(self):
        return pd.DataFrame(self.columns_dict)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

import numpy as np

from .common import (
    make_prefix_suffix_pairs,
    nucleotide_to_index,
    index_to_nucleotide,
)

def nucleotide_counts(partitioned_read_sequences):
    """
    Count the number of times {A, C, T, G} occur at each position to the
    left and right of the variant.

    Parameters
    ----------
    partitioned_read_sequences : list of tuples
        Each tuple has three elements:
            - sequence before mutant nucleotides
            - mutant nucleotides
            - sequence after mutant nucleotides

    Returns a tuple with the following elements:
        - a matrix with four rows and as many columns as the sum of the longest
          prefix preceding the variant, the longest suffix after the variant and
          the number of variant nucleotids.
        - the column indices for the variant nucleotides
    """
    variant_seq, prefix_suffix_pairs = make_prefix_suffix_pairs(
        partitioned_read_sequences)
    n_reads = len(prefix_suffix_pairs)
    max_prefix_length = max(len(p) for (p, _) in prefix_suffix_pairs)
    max_suffix_length = max(len(s) for (_, s) in prefix_suffix_pairs)
    n_variant_nucleotides = len(variant_seq)

    n_cols = max_prefix_length + max_suffix_length + n_variant_nucleotides

    counts = np.zeros((4, n_cols), dtype=int)

    variant_column_indices = []

    # first fill in the variant nucleotide counts, since they'll
    # be invariant across all the supporting reads
    for i, nucleotide in enumerate(variant_seq):
        variant_col_idx = max_prefix_length + i
        variant_column_indices.append(variant_col_idx)
        row_idx = nucleotide_to_index[nucleotide_to_index]
        counts[row_idx, variant_col_idx] = n_reads

    for p, s in prefix_suffix_pairs:
        for i, prefix_col_idx in enumerate(range(
                max_prefix_length - len(p),
                max_prefix_length)):
            row_idx = nucleotide_to_index[p[i]]
            counts[row_idx, prefix_col_idx] += 1
        for i, suffix_col_idx in enumerate(range(
                max_prefix_length + n_variant_nucleotides,
                max_prefix_length + n_variant_nucleotides + len(s))):
            row_idx = nucleotide_to_index[s[i]]
            counts[row_idx, suffix_col_idx] += 1
    return counts, variant_column_indices

def most_common_nucleotides(partitioned_read_sequences):
    """
    Find the most common nucleotide at each offset to the left and
    right of a variant.

    Parameters
    ----------
    partitioned_read_sequences : list of tuples
        Each tuple has three elements:
            - sequence before mutant nucleotides
            - mutant nucleotides
            - sequence after mutant nucleotides

    Returns a tuple with the following elements:
        - nucleotide sequence from most common nucleotide at each offset
           relative to the variant
        - an array of counts indicating how many reads supported this nucleotide
        - an array of counts for all the *other* nucleotides at that position
    """
    counts, variant_column_indices = nucleotide_counts(
        partitioned_read_sequences)
    max_count_per_column = counts.max(axis=0)

    assert len(max_count_per_column) == counts.shape[1]
    max_nucleotide_index_per_column = np.argmax(counts, axis=0)
    assert len(max_nucleotide_index_per_column) == counts.shape[1]
    nucleotides = [
        index_to_nucleotide[idx]
        for idx in max_nucleotide_index_per_column
    ]
    other_nucleotide_counts = counts.sum(axis=0) - max_count_per_column
    return "".join(nucleotides), max_count_per_column, other_nucleotide_counts



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import


from isovar.translation import (
    translate_variants,
    compute_offset_to_first_complete_codon,
)
from nose.tools import eq_

from testing_helpers import load_bam, load_vcf


def test_compute_offset_to_first_complete_codon_no_trimming():
    # if nothing gets trimmed from the reference sequence, then
    # the offset to the first codon shouldn't change
    eq_(
        compute_offset_to_first_complete_codon(
            offset_to_first_complete_reference_codon=0,
            n_trimmed_from_reference_sequence=0),
        0)
    eq_(
        compute_offset_to_first_complete_codon(
            offset_to_first_complete_reference_codon=5,
            n_trimmed_from_reference_sequence=0),
        5)

def test_compute_offset_to_first_complete_codon_trimming_before_codon():
    # if the number of reference bases trimmed from the reference sequence
    # occurs before the reference codon, then it should decrease the
    # offset by the amount trimmed
    eq_(
        compute_offset_to_first_complete_codon(
            offset_to_first_complete_reference_codon=7,
            n_trimmed_from_reference_sequence=2),
        5)
    eq_(
        compute_offset_to_first_complete_codon(
            offset_to_first_complete_reference_codon=7,
            n_trimmed_from_reference_sequence=7),
        0)

def test_compute_offset_to_first_complete_codon_trimming_after_codon():
    # if the number of reference bases trimmed from the reference sequence
    # occurs after the reference codon, then it needs to be rounded up the
    # next multiple of three
    eq_(
        compute_offset_to_first_complete_codon(
            offset_to_first_complete_reference_codon=7,
            n_trimmed_from_reference_sequence=8),
        2)
    eq_(
        compute_offset_to_first_complete_codon(
            offset_to_first_complete_reference_codon=7,
            n_trimmed_from_reference_sequence=10),
        0)

"""
def test_determine_reading_frame_for_variant_sequence():
    reference_context = None
    variant_sequence = None
    variant_sequence_in_reading_frame = \
        determine_reading_frame_for_variant_sequence(
            variant_sequence=variant_sequence,
            reference_context=reference_context)
"""

def test_translate_variant_collection():
    variants = load_vcf("data/b16.f10/b16.vcf")
    samfile = load_bam("data/b16.f10/b16.combined.sorted.bam")
    result = list(translate_variants(variants, samfile))
    assert len(result) == 4, result

if __name__ == "__main__":
    test_translate_variant_collection()



#!/usr/bin/env python

# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

import argparse

import varcode

from isovar.reference_context import variants_to_reference_contexts_dataframe
from isovar.default_parameters import CDNA_CONTEXT_SIZE
from isovar.args import add_somatic_vcf_args

parser = argparse.ArgumentParser()
parser = add_somatic_vcf_args(parser)

parser.add_argument(
    "--context-size",
    default=CDNA_CONTEXT_SIZE,
    type=int)

parser.add_argument(
    "--output",
    default="isovar-reference-contexts-result.csv",
    help="Name of output CSV")

if __name__ == "__main__":
    args = parser.parse_args()
    variants = varcode.load_vcf(args.vcf, genome=args.genome)
    reference_contexts_df = variants_to_reference_contexts_dataframe(
        variants=variants,
        context_size=args.context_size)
    print(reference_contexts_df)
    reference_contexts_df.to_csv(args.output)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from isovar.nucleotide_counts import most_common_nucleotides
from isovar.variant_read import gather_reads_for_single_variant

from varcode import Variant
from pyensembl import ensembl_grch38
from nose.tools import eq_

from testing_helpers import load_bam


def test_most_common_nucleotides_for_chr12_deletion():
    samfile = load_bam("data/cancer-wgs-primary.chr12.bam")
    chromosome = "chr12"
    base1_location = 70091490
    ref = "TTGTAGATGCTGCCTCTCC"
    alt = ""
    variant = Variant(
        chromosome,
        base1_location,
        ref,
        alt,
        ensembl=ensembl_grch38)
    variant_reads = gather_reads_for_single_variant(
        samfile=samfile,
        chromosome=chromosome,
        variant=variant)
    consensus_sequence, chosen_counts, other_counts = most_common_nucleotides(
        variant_reads)
    print(chosen_counts)
    print(other_counts)
    eq_(len(chosen_counts), len(consensus_sequence))
    eq_(len(other_counts), len(consensus_sequence))
    assert other_counts.sum() < chosen_counts.sum(), \
        "Counts for alternate nucleotides should not exceed the chosen sequence"

    number_matching_reads = 0
    for variant_read in variant_reads:
        full_seq = variant_read.prefix + variant_read.alt + variant_read.suffix
        number_matching_reads += (full_seq in consensus_sequence)
    fraction_matching_reads = number_matching_reads / float(len(variant_reads))
    print("Fraction matching reads is %d/%d = %f" % (
        number_matching_reads, len(variant_reads), fraction_matching_reads))
    assert fraction_matching_reads > 0.5, \
        "Expected majority of reads to match consensus sequence"

if __name__ == "__main__":
    test_most_common_nucleotides_for_chr12_deletion()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import
from collections import namedtuple
import logging


from .common import group_unique_sequences, get_variant_nucleotides
from .variant_read import variant_reads_generator
from .default_parameters import (
    MIN_READS_SUPPORTING_VARIANT_CDNA_SEQUENCE,
    VARIANT_CDNA_SEQUENCE_LENGTH,
    MIN_READ_MAPPING_QUALITY,
)
from .dataframe_builder import DataFrameBuilder

VariantSequence = namedtuple(
    "VariantSequence",
    [
        "prefix",
        "alt",
        "suffix",
        # combined sequence cached for convenience, so we don't have to
        # repeatedly concatenate prefix + variant_nucleotides + suffix
        "full_sequence",
        "read_names",
        "read_count",
    ]
)

def sort_key_decreasing_read_count(variant_sequence):
    return -variant_sequence.read_count

def variant_reads_to_sequences(
        variant_reads,
        max_nucleotides_before_variant=None,
        max_nucleotides_after_variant=None,
        min_sequence_length=None,
        min_reads_per_sequence=MIN_READS_SUPPORTING_VARIANT_CDNA_SEQUENCE):
    """
    Parameters
    ----------
    variant_reads : list of VariantRead objects
        Each variant read should have the following fields:
            - prefix : str
            - variant : str
            - suffix : str
            - name : str

    max_nucleotides_before_variant : int, optional
        Number of nucleotides left of a variant. If omitted then
        use all prefix nucleotides.

    max_nucleotides_after_variant : int, optional
        Number of nucleotides right of a variant. If omitted then
        use all suffix nucleotides.

    min_sequence_length : int, optional
        Minimum length of detected sequence

    min_reads_per_sequence : int, optional
        Drop sequences which don't at least have this number of fully spanning
        reads.

    Returns a collection of VariantSequence objects
    """
    # just in case variant_reads is a generator, convert it to a list
    variant_reads = list(variant_reads)

    if len(variant_reads) == 0:
        return []

    # Get all unique sequences from reads spanning the
    # variant locus. This will include partial sequences
    # due to reads starting in the middle of the context,
    # we'll use these only to compute partial support for a
    # full length sequence
    unique_sequence_groups = group_unique_sequences(
        variant_reads,
        max_prefix_size=max_nucleotides_before_variant,
        max_suffix_size=max_nucleotides_after_variant)

    variant_seq = get_variant_nucleotides(variant_reads)
    variant_len = len(variant_seq)

    if min_sequence_length is None or min_sequence_length == 0:
        # if not specified, then use the longest sequence available
        # which will be at most twice the context size
        min_sequence_length = variant_len + max(
            len(prefix) + len(suffix)
            for (prefix, suffix) in unique_sequence_groups.keys())

    variant_sequences = [
        VariantSequence(
            prefix=prefix,
            alt=variant_seq,
            suffix=suffix,
            full_sequence=prefix + variant_seq + suffix,
            read_names=read_names,
            read_count=len(read_names))
        for ((prefix, suffix), read_names)
        in unique_sequence_groups.items()
    ]
    n_total = len(variant_sequences)

    variant_sequences = [
        x for x in variant_sequences
        if len(x.full_sequence) >= min_sequence_length and
        len(x.read_names) >= min_reads_per_sequence
    ]

    n_dropped = n_total - len(variant_sequences)
    logging.info("Dropped %d/%d variant sequences" % (n_dropped, n_total))

    # sort VariantSequence objects by decreasing order of supporting read
    # counts
    variant_sequences.sort(key=sort_key_decreasing_read_count)
    return variant_sequences

def variant_sequences_generator(
        variants,
        samfile,
        sequence_length=VARIANT_CDNA_SEQUENCE_LENGTH,
        min_reads=MIN_READS_SUPPORTING_VARIANT_CDNA_SEQUENCE,
        min_mapping_quality=MIN_READ_MAPPING_QUALITY):
    """
    For each variant, collect all possible sequence contexts around the
    variant which are spanned by at least min_reads.

    Parameters
    ----------
    variants : varcode.VariantCollection
        Variants for which we're trying to construct context sequences

    samfile : pysam.AlignmentFile
        BAM or SAM file containing RNA reads

    sequence_length : int
        Desired sequence length, including variant nucleotides

    min_reads : int
        Minimum number of reads supporting a particular sequence

    min_mapping_quality : int
        Minimum MAPQ value before a read gets ignored

    Generator that yields tuples with the following fields:
        - Variant
        - list of VariantSequence objects
    """
    for variant, variant_reads in variant_reads_generator(
            variants=variants,
            samfile=samfile,
            min_mapping_quality=min_mapping_quality):
        # the number of context nucleotides on either side of the variant
        # is half the desired length (minus the number of variant nucleotides)
        n_alt = len(variant.alt)
        n_surrounding_nucleotides = sequence_length - n_alt
        max_nucleotides_after_variant = n_surrounding_nucleotides // 2
        # if the number of nucleotides we need isn't divisible by 2 then
        # prefer to have one more *before* the variant since we need the
        # prefix sequence to match against reference transcripts
        max_nucleotides_before_variant = (
            n_surrounding_nucleotides - max_nucleotides_after_variant)
        logging.info(
            "Looking at %dnt before and %dnt after variant %s" % (
                max_nucleotides_before_variant,
                max_nucleotides_after_variant,
                variant))

        variant_sequences = variant_reads_to_sequences(
            variant_reads,
            max_nucleotides_before_variant=max_nucleotides_before_variant,
            max_nucleotides_after_variant=max_nucleotides_after_variant,
            min_reads_per_sequence=min_reads)

        yield variant, variant_sequences

def variant_sequences_dataframe(
        variants,
        samfile,
        sequence_length=VARIANT_CDNA_SEQUENCE_LENGTH,
        min_reads=MIN_READS_SUPPORTING_VARIANT_CDNA_SEQUENCE,
        min_mapping_quality=MIN_READ_MAPPING_QUALITY):
    """
    Creates a dataframe of all detected cDNA sequences for the given variant
    collection and alignment file.

    Parameters
    ----------
    variants : varcode.VariantCollection
        Look for sequences containing these variants

    samfile : pysam.AlignmentFile
        Reads from which surrounding sequences are detected

    sequence_length : int
        Desired sequence length, including variant nucleotides

    min_reads : int
        Minimum number of reads supporting

    min_mapping_quality : int
        Minimum MAPQ value before a read gets ignored

    Returns pandas.DataFrame
    """
    df_builder = DataFrameBuilder(VariantSequence)
    for variant, variant_sequences, variant_reads in variant_sequences_generator(
            variants=variants,
            samfile=samfile,
            sequence_length=sequence_length,
            min_reads=min_reads,
            min_mapping_quality=min_mapping_quality):
        for variant_sequence in variant_sequences:
            df_builder.add(variant, variant_sequence)
    return df_builder.to_dataframe()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import
from collections import namedtuple, OrderedDict, defaultdict
import logging

from skbio import DNA

from .effect_prediction import reference_transcripts_for_variant
from .variant_helpers import interbase_range_affected_by_variant_on_transcript
from .dataframe_builder import DataFrameBuilder

##########################
#
# SequenceKey
# -----------
#
# Used to identify and group the distinct sequences occurring  on a set of
# transcripts overlapping a variant locus
#
##########################

SequenceKey = namedtuple(
    "SequenceKey", [
        "strand",
        "sequence_before_variant_locus",
        "sequence_at_variant_locus",
        "sequence_after_variant_locus"
    ]
)


##########################
#
# SequenceKeyWithReadingFrame
# ---------------------------
#
# Includes all the fields of a SequenceKey, but also includes a reading frame
# at the start of the reference sequence.
#
#
##########################

SequenceKeyWithReadingFrame = namedtuple(
    "ReferenceContext", SequenceKey._fields + (
        # if the reference context includes the 5' UTR then
        # this is the offset to the start codon, otherwise it's the
        # offset needed to get the first base of a codon
        "offset_to_first_complete_codon",
        # does this context overlap a start codon?
        "overlaps_start_codon",
        # does this context contain the whole trinucleotide start codon?
        "contains_start_codon",
        # does this context contain any UTR bases?
        "contains_five_prime_utr",
        # translation of complete codons in the reference context
        # before the variant
        "amino_acids_before_variant",
    )
)


##########################
#
# ReferenceContext
# ----------------
#
# Includes all the fields of SequenceKeyWithReadingFrame in addition to which
# variant we're examining and all transcripts overlapping that variant
# which produced this particular sequence context and reading frame.
#
##########################

ReferenceContext = namedtuple(
    "ReferenceContext",
    SequenceKeyWithReadingFrame._fields + ("variant", "transcripts")
)

def variant_matches_reference_sequence(variant, ref_seq_on_transcript, strand):
    """
    Make sure that reference nucleotides we expect to see on the reference
    transcript from a variant are the same ones we encounter.
    """
    if strand == "-":
        ref_seq_on_transcript = str(
            DNA(ref_seq_on_transcript).reverse_complement())

    return ref_seq_on_transcript == variant.ref

def sequence_key_for_variant_on_transcript(variant, transcript, context_size):
    """
    Extracts the reference sequence around a variant locus on a particular
    transcript.

    Parameters
    ----------
    variant : varcode.Variant

    transcript : pyensembl.Transcript

    context_size : int

    Returns SequenceKey object with the following fields:
        - strand
        - sequence_before_variant_locus
        - sequence_at_variant_locus
        - sequence_after_variant_locus

    Can also return None if Transcript lacks sufficiently long sequence
    """

    full_sequence = transcript.sequence

    if full_sequence is None:
        logging.warn(
            "Expected transcript %s (overlapping %s) to have sequence" % (
                transcript,
                variant))
        return None

    if len(full_sequence) < 6:
        # need at least 6 nucleotides for a start and stop codon
        logging.warn(
            "Sequence of %s (overlapping %s) too short: %d" % (
                transcript,
                variant,
                len(full_sequence)))
        return None

    # get the interbase range of offsets which capture all reference
    # bases modified by the variant
    variant_start_offset, variant_end_offset = \
        interbase_range_affected_by_variant_on_transcript(
            variant=variant,
            transcript=transcript)

    logging.info("Interbase offset range on %s for variant %s = %d:%d" % (
        transcript,
        variant,
        variant_start_offset,
        variant_end_offset))

    prefix = full_sequence[
        max(0, variant_start_offset - context_size):
        variant_start_offset]

    suffix = full_sequence[
        variant_end_offset:
        variant_end_offset + context_size]

    ref_nucleotides_at_variant = full_sequence[
        variant_start_offset:variant_end_offset]
    if not variant_matches_reference_sequence(
            variant=variant,
            strand=transcript.strand,
            ref_seq_on_transcript=ref_nucleotides_at_variant):
        # TODO: once we're more confident about other logic in isovar,
        # change this to a warning and return None to allow for modest
        # differences between reference sequence patches, since
        # GRCh38.p1 may differ at some positions from GRCh38.p5
        raise ValueError(
            "Wrong reference sequence for variant %s on transcript %s" % (
                variant,
                transcript))
    return SequenceKey(
        strand=transcript.strand,
        sequence_before_variant_locus=prefix,
        sequence_at_variant_locus=ref_nucleotides_at_variant,
        sequence_after_variant_locus=suffix)


def reading_frame_to_offset(reading_frame_at_start_of_sequence):
    """
    Given a reading frame (how many nucleotides into a codon) at the
    start of a cDNA sequence, return the number of nucleotides which need
    to be trimmed to start on a complete codon.

    Parameters
    ----------

    reading_frame_at_start_of_sequence : int

    Returns an int
    """
    if reading_frame_at_start_of_sequence < 0:
        raise ValueError("Reading frame can't be negative: %d" % (
            reading_frame_at_start_of_sequence,))
    elif reading_frame_at_start_of_sequence > 2:
        raise ValueError("Reading frame must be within 0 and 2, not %d" % (
            reading_frame_at_start_of_sequence,))
    # If we're 1 nucleotide into the codon then we need to shift
    # over two more to restore the ORF. Likewise, if we're 2 nucleotides in
    # then we have to shift over one more.
    return (3 - reading_frame_at_start_of_sequence) % 3


def sequence_key_with_reading_frame_for_variant_on_transcript(
        variant,
        transcript,
        context_size):
    """
    Extracts the reference sequence around a variant locus on a particular
    transcript and determines the reading frame at the start of that
    sequence context.

    Parameters
    ----------
    variant : varcode.Variant

    transcript : pyensembl.Transcript

    context_size : int

    Returns SequenceKeyWithReadingFrame object or None if Transcript lacks
    coding sequence, protein sequence or annotated start/stop codons.
    """
    if not transcript.contains_start_codon:
        logging.info(
            "Expected transcript %s for variant %s to have start codon" % (
                transcript,
                variant))
        return None

    if not transcript.contains_stop_codon:
        logging.info(
            "Expected transcript %s for variant %s to have stop codon" % (
                transcript,
                variant))
        return None

    if not transcript.protein_sequence:
        logging.info(
            "Expected transript %s for variant %s to have protein sequence" % (
                transcript,
                variant))
        return None

    sequence_key = sequence_key_for_variant_on_transcript(
        variant=variant,
        transcript=transcript,
        context_size=context_size)

    if sequence_key is None:
        logging.info("No sequence key for variant %s on transcript %s" % (
            variant,
            transcript))
        return None

    # get the interbase range of offsets which capture all reference
    # bases modified by the variant
    variant_start_offset, variant_end_offset = \
        interbase_range_affected_by_variant_on_transcript(
            variant=variant,
            transcript=transcript)

    start_codon_idx = min(transcript.start_codon_spliced_offsets)

    # skip any variants which occur in the 5' UTR or overlap the start codon
    # since
    #   (1) UTR variants have unclear coding effects and
    #   (2) start-loss variants may result in a new start codon / reading
    #       frame but we're not sure which one!
    if variant_start_offset < start_codon_idx + 3:
        logging.info(
            "Skipping transcript %s for variant %s, must be after start codon" % (
                transcript,
                variant))
        return None

    stop_codon_idx = min(transcript.stop_codon_spliced_offsets)

    # skip variants which affect the 3' UTR of the transcript since
    # they don't have obvious coding effects on the protein sequence
    if variant_start_offset >= stop_codon_idx + 3:
        logging.info(
            "Skipping transcript %s for variant %s, occurs in 3' UTR" % (
                transcript,
                variant))
        return None

    n_prefix = len(sequence_key.sequence_before_variant_locus)
    prefix_start_idx = variant_start_offset - n_prefix
    n_bases_between_start_and_variant = variant_start_offset - start_codon_idx
    n_full_codons_before_variant = n_bases_between_start_and_variant // 3

    # if the sequence before the variant contains more bases than the
    # distance to the start codon, then by definition it must contain
    # some untranslated bases
    contains_five_prime_utr = (n_prefix > n_bases_between_start_and_variant)
    # allows for the possibility that the first base in the sequence might
    # be the first nucleotide of the start codon
    contains_start_codon = (n_prefix >= n_bases_between_start_and_variant)
    # the sequence context might only include the 2nd or 3rd bases of
    # the start codon
    overlaps_start_codon = (n_prefix > n_bases_between_start_and_variant - 3)

    if contains_start_codon:
        offset_to_first_complete_codon = start_codon_idx - prefix_start_idx
        amino_acids_before_variant = transcript.protein_sequence[:n_full_codons_before_variant]
    else:
        reading_frame = (prefix_start_idx - start_codon_idx) % 3
        offset_to_first_complete_codon = reading_frame_to_offset(reading_frame)
        n_codons_in_prefix = (n_prefix - offset_to_first_complete_codon) // 3
        amino_acids_before_variant = transcript.protein_sequence[
            n_full_codons_before_variant - n_codons_in_prefix:
            n_full_codons_before_variant]
    return SequenceKeyWithReadingFrame(
        strand=sequence_key.strand,
        sequence_before_variant_locus=sequence_key.sequence_before_variant_locus,
        sequence_at_variant_locus=sequence_key.sequence_at_variant_locus,
        sequence_after_variant_locus=sequence_key.sequence_after_variant_locus,
        offset_to_first_complete_codon=offset_to_first_complete_codon,
        contains_start_codon=contains_start_codon,
        overlaps_start_codon=overlaps_start_codon,
        contains_five_prime_utr=contains_five_prime_utr,
        amino_acids_before_variant=amino_acids_before_variant)

def sort_key_decreasing_max_length_transcript_cds(reference_context):
    """
    Used to sort a sequence of ReferenceContext objects by the longest CDS
    in each context's list of transcripts.
    """
    return -max(len(t.coding_sequence) for t in reference_context.transcripts)

def reference_contexts_for_variant(
        variant,
        context_size,
        transcript_id_whitelist=None):
    """
    variant : varcode.Variant

    context_size : int
        Max of nucleotides to include to the left and right of the variant
        in the context sequence.

    transcript_id_whitelist : set, optional
        If given, then only consider transcripts whose IDs are in this set.

    Returns list of ReferenceContext objects, sorted by maximum length of
    coding sequence of any supporting transcripts.
    """
    overlapping_transcripts = reference_transcripts_for_variant(
        variant=variant,
        transcript_id_whitelist=transcript_id_whitelist)

    # dictionary mapping SequenceKeyWithReadingFrame keys to list of
    # transcript objects
    sequence_groups = defaultdict(list)

    for transcript in overlapping_transcripts:
        sequence_key_with_reading_frame = \
            sequence_key_with_reading_frame_for_variant_on_transcript(
                variant=variant,
                transcript=transcript,
                context_size=context_size)
        if sequence_key_with_reading_frame is not None:
            sequence_groups[sequence_key_with_reading_frame].append(transcript)

    reference_contexts = [
        ReferenceContext(
            strand=key.strand,
            sequence_before_variant_locus=key.sequence_before_variant_locus,
            sequence_at_variant_locus=key.sequence_at_variant_locus,
            sequence_after_variant_locus=key.sequence_after_variant_locus,
            offset_to_first_complete_codon=key.offset_to_first_complete_codon,
            contains_start_codon=key.contains_start_codon,
            overlaps_start_codon=key.overlaps_start_codon,
            contains_five_prime_utr=key.contains_five_prime_utr,
            amino_acids_before_variant=key.amino_acids_before_variant,
            variant=variant,
            transcripts=matching_transcripts)
        for (key, matching_transcripts) in sequence_groups.items()
    ]
    reference_contexts.sort(key=sort_key_decreasing_max_length_transcript_cds)
    return reference_contexts

def reference_contexts_for_variants(
        variants,
        context_size,
        transcript_id_whitelist=None):
    """
    Extract a set of reference contexts for each variant in the collection.

    Parameters
    ----------
    variants : varcode.VariantCollection

    context_size : int
        Max of nucleotides to include to the left and right of the variant
        in the context sequence.

    transcript_id_whitelist : set, optional
        If given, then only consider transcripts whose IDs are in this set.

    Returns a dictionary from variants to lists of ReferenceContext objects,
    sorted by max coding sequence length of any transcript.
    """
    result = OrderedDict()
    for variant in variants:
        result[variant] = reference_contexts_for_variant(
            variant=variant,
            context_size=context_size,
            transcript_id_whitelist=transcript_id_whitelist)
    return result

def variants_to_reference_contexts_dataframe(
        variants,
        context_size,
        transcript_id_whitelist=None):
    """
    Given a collection of variants, find all reference sequence contexts
    around each variant.

    Parameters
    ----------
    variants : varcode.VariantCollection

    context_size : int
        Max of nucleotides to include to the left and right of the variant
        in the context sequence.

    transcript_id_whitelist : set, optional
        If given, then only consider transcripts whose IDs are in this set.

    Returns a DataFrame with {"chr", "pos", "ref", "alt"} columns for variants,
    as well as all the fields of ReferenceContext.
    """

    df_builder = DataFrameBuilder(
        ReferenceContext,
        exclude=["variant"],
        converters=dict(transcripts=lambda ts: ";".join(t.name for t in ts)))
    for variant, reference_contexts in reference_contexts_for_variants(
            variants=variants,
            context_size=context_size,
            transcript_id_whitelist=transcript_id_whitelist).items():
        for reference_context in reference_contexts:
            df_builder.add(variant, reference_context)
    return df_builder.to_dataframe()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from varcode import Variant
from isovar.variant_read import gather_reads_for_single_variant, VariantRead

from mock_read_data import DummySamFile, make_read
from testing_helpers import assert_equal_fields

def test_partitioned_read_sequences_snv():
    """
    test_partitioned_read_sequences_snv : Test that read gets correctly
    partitioned for chr1:4 T>G where the sequence for chr1 is assumed
    to be "ACCTTG"
    """
    # chr1_seq = "ACCTTG"
    chromosome = "chromosome"
    location = 4
    ref = "T"
    alt = "G"

    variant = Variant(
        chromosome, location, ref, alt, normalize_contig_name=False)

    read = make_read(seq="ACCGTG", cigar="6M", mdtag="3G2")

    samfile = DummySamFile(reads=[read])
    variant_reads = gather_reads_for_single_variant(
        samfile=samfile,
        chromosome=chromosome,
        variant=variant)
    print(variant_reads)
    assert len(variant_reads) == 1
    variant_read = variant_reads[0]
    expected = VariantRead(
        name=read.qname,
        prefix="ACC",
        alt="G",
        suffix="TG")
    assert_equal_fields(variant_read, expected)

def test_partitioned_read_sequences_insertion():
    """
    test_partitioned_read_sequences_insertion : Test that read gets correctly
    partitioned for chr1:4 T>TG
    where the sequence for chr1 is assumed to be "ACCTTG"
    and the variant sequence is "ACCTGTG"
    """
    # chr1_seq = "ACCTTG"
    chromosome = "chromosome"
    location = 4
    ref = "T"
    alt = "TG"
    variant = Variant(
        chromosome, location, ref, alt, normalize_contig_name=False)

    read = make_read(seq=b"ACCTGTG", cigar="4M1I2M", mdtag="6")

    samfile = DummySamFile(reads=[read])
    variant_reads = gather_reads_for_single_variant(
        samfile=samfile,
        chromosome=chromosome,
        variant=variant)
    print(variant_reads)
    assert len(variant_reads) == 1
    variant_read = variant_reads[0]
    expected = VariantRead(
        name=read.qname,
        prefix="ACCT",
        alt="G",
        suffix="TG")
    assert_equal_fields(variant_read, expected)

def test_partitioned_read_sequences_deletion():
    """
    test_partitioned_read_sequences_deletion : Test that read gets correctly
    partitioned for chr1:4 TT>T where the sequence for chr1 is assumed to
    be "ACCTTG"
    """
    # chr1_seq = "ACCTTG"
    chromosome = "chromosome"
    location = 4
    ref = "TT"
    alt = "T"
    variant = Variant(
        chromosome, location, ref, alt, normalize_contig_name=False)

    read = make_read(seq="ACCTG", cigar="4M1D1M", mdtag="4^T1")

    samfile = DummySamFile(reads=[read])
    variant_reads = gather_reads_for_single_variant(
        samfile=samfile,
        chromosome=chromosome,
        variant=variant)
    print(variant_reads)
    assert len(variant_reads) == 1
    variant_read = variant_reads[0]
    expected = VariantRead(
        name=read.qname,
        prefix="ACCT",
        alt="",
        suffix="G")
    assert_equal_fields(variant_read, expected)



from isovar.args import (
    add_somatic_vcf_args,
    add_rna_args,
    add_reference_context_args,
    add_protein_sequence_args,
)
from argparse import ArgumentParser
from nose.tools import eq_

def test_extend_parser():
    parser = ArgumentParser()
    fns = [
        add_somatic_vcf_args,
        add_rna_args,
        add_reference_context_args,
        add_protein_sequence_args,
    ]
    for fn in fns:
        parser = fn(parser)
    args = parser.parse_args([
        "--vcf", "ABC.vcf",
        "--bam", "xyz.bam"])
    eq_(args.vcf, "ABC.vcf")
    eq_(args.bam, "xyz.bam")


if __name__ == "__main__":
    test_extend_parser()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helper functions for building ArgumentParser instances. Keeps us from
having to copy the same arguments across many scripts.
"""

from __future__ import print_function, division, absolute_import

from .default_parameters import (
    MIN_READS_SUPPORTING_VARIANT_CDNA_SEQUENCE,
    MAX_REFERENCE_TRANSCRIPT_MISMATCHES,
    MIN_TRANSCRIPT_PREFIX_LENGTH,
    MIN_READ_MAPPING_QUALITY,
    PROTEIN_SEQUENCE_LENGTH,
    MAX_PROTEIN_SEQUENCES_PER_VARIANT,
)

def add_somatic_vcf_args(
        parser,
        vcf_arg="--vcf",
        vcf_required=True,
        genome_arg="--genome",
        genome_required=False):
    """
    Extends an ArgumentParser instance with the following commandline arguments:
        --vcf
        --genome
    """
    parser.add_argument(
        vcf_arg,
        required=vcf_required,
        help="Path to VCF file containing somatic variants")

    parser.add_argument(
        genome_arg,
        default=None,
        required=genome_required,
        help="Name of reference genome for VCF of somatic variants")
    return parser

def add_rna_args(
        parser,
        rna_bam_arg="--bam",
        rna_bam_required=True,
        min_rna_reads_arg="--min-reads",
        min_rna_reads_default=MIN_READS_SUPPORTING_VARIANT_CDNA_SEQUENCE,
        min_mapping_quality_arg="--min-mapping-quality",
        min_mapping_quality_default=MIN_READ_MAPPING_QUALITY):
    """
    Extends an ArgumentParser instance with the following commandline arguments:
        --bam
        --min-reads
        --min-mapping-quality
    """
    parser.add_argument(
        rna_bam_arg,
        required=rna_bam_required,
        help="BAM file containing RNAseq reads")

    parser.add_argument(
        min_rna_reads_arg,
        type=int,
        default=min_rna_reads_default,
        help="Minimum number of reads supporting a variant sequence")

    parser.add_argument(
        "--min-mapping-quality",
        type=int,
        default=min_mapping_quality_default,
        help="Minimum MAPQ value to allow for a read")
    return parser

def add_reference_context_args(
        parser,
        max_reference_transcript_mismatches_arg="--max-reference-transcript-mismatches",
        max_reference_transcript_mismatches_default=MAX_REFERENCE_TRANSCRIPT_MISMATCHES,
        min_transcript_prefix_length_arg="--min-transcript-prefix-length",
        min_transcript_prefix_length_default=MIN_TRANSCRIPT_PREFIX_LENGTH):
    """
    Extends an ArgumentParser instance with the following commandline arguments:
        --max-reference-transcript-mismatches
        --min-transcript-prefix-length
    """
    parser.add_argument(
        max_reference_transcript_mismatches_arg,
        type=int,
        default=max_reference_transcript_mismatches_default,
        help=(
            "Maximum number of mismatches between variant sequence"
            " reference sequence before a candidate reading frame is ignored."))

    parser.add_argument(
        min_transcript_prefix_length_arg,
        type=int,
        default=min_transcript_prefix_length_default,
        help=(
            "Number of nucleotides before the variant we try to match against "
            "a reference transcript. Values greater than zero exclude variants "
            "near the start codon of transcrPROTEIN_SEQUENCE_LEGNTHipts without 5' UTRs."))
    return parser

def add_protein_sequence_args(
        parser,
        protein_sequence_length_arg="--protein-sequence-length",
        protein_sequence_length_default=PROTEIN_SEQUENCE_LENGTH,
        max_protein_sequences_per_variant_arg="--max-protein-sequences-per-variant",
        max_protein_sequences_per_variant_default=MAX_PROTEIN_SEQUENCES_PER_VARIANT):
    """
    Extends an ArgumentParser instance with the following args:
        --protein-sequence-length
        --max-protein-sequences-per-variant
    """
    parser.add_argument(
        protein_sequence_length_arg,
        default=protein_sequence_length_default,
        type=int)
    parser.add_argument(
        max_protein_sequences_per_variant_arg,
        type=int,
        default=max_protein_sequences_per_variant_default)
    return parser



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from nose.tools import eq_
from isovar.translation import Translation
from isovar.protein_sequence import (
    ProteinSequence,
    sort_protein_sequences,
    variants_to_protein_sequences_dataframe
)
from varcode import VariantCollection

from testing_helpers import load_bam, load_vcf


# fields of a ProteinSequence:
#   translations
#   supporting_variant_reads
#   total_variant_reads
#   supporting_transcripts
#   total_transcripts
#   gene

def make_dummy_translation(
        amino_acids="MKHW",  # ATG=M|AAA=K|CAC=H|TGG=W
        cdna_sequence="CCCATGAAACACTGGTAG",
        offset_to_first_complete_codon=3,
        variant_cdna_interval_start=8,  # assuming variant was AAC>AAA
        variant_cdna_interval_end=9,
        variant_aa_interval_start=1,
        variant_aa_interval_end=2,
        number_mismatches=1):
    return Translation(
        variant_sequence_in_reading_frame=None,
        cdna_sequence=cdna_sequence,
        offset_to_first_complete_codon=offset_to_first_complete_codon,
        variant_cdna_interval_start=variant_cdna_interval_start,
        variant_cdna_interval_end=variant_cdna_interval_end,
        reference_cdna_sequence_before_variant=cdna_sequence[:variant_cdna_interval_start],
        number_mismatches=number_mismatches,
        amino_acids=amino_acids,
        variant_aa_interval_start=variant_aa_interval_start,
        variant_aa_interval_end=variant_aa_interval_end,
        frameshift=False,
        ends_with_stop_codon=False,
        variant_sequence=None,
        reference_context=None)

def make_dummy_protein_sequence(
        n_supporting_variant_reads,
        n_supporting_variant_sequences,
        n_supporting_reference_transcripts,
        n_total_variant_sequences=None,
        n_total_variant_reads=None,
        n_total_reference_transcripts=None,
        gene=["TP53"],
        amino_acids="MKHW",  # ATG=M|AAA=K|CAC=H|TGG=W
        cdna_sequence="CCCATGAAACACTGGTAG",
        variant_cdna_interval_start=8,  # assuming variant was AAC>AAA
        variant_cdna_interval_end=9,
        variant_aa_interval_start=1,
        variant_aa_interval_end=2,
        number_mismatches=1):
    """
    Creates ProteinSequence object with None filled in for most fields
    """
    if n_total_variant_reads is None:
        n_total_variant_reads = n_supporting_variant_reads

    if n_total_variant_sequences is None:
        n_total_variant_sequences = n_supporting_variant_sequences

    if n_total_reference_transcripts is None:
        n_total_reference_transcripts = n_total_reference_transcripts

    assert n_supporting_variant_sequences <= n_supporting_variant_reads
    assert n_supporting_variant_sequences <= n_total_variant_sequences
    assert n_supporting_reference_transcripts <= n_total_reference_transcripts

    n_translations = n_total_reference_transcripts * n_total_variant_sequences

    translation = make_dummy_translation()

    return ProteinSequence(
        translations=[translation] * n_translations,
        supporting_variant_reads=[None] * n_supporting_variant_reads,
        total_variant_reads=n_total_variant_reads,
        supporting_transcripts=[None] * n_supporting_reference_transcripts,
        total_transcripts=n_total_reference_transcripts,
        gene=gene,
        amino_acids=amino_acids,
        variant_aa_interval_start=variant_aa_interval_start,
        variant_aa_interval_end=variant_aa_interval_end,
        ends_with_stop_codon=translation.ends_with_stop_codon,
        frameshift=translation.frameshift)

def test_sort_protein_sequences():
    protseq_most_reads = make_dummy_protein_sequence(
        n_supporting_variant_reads=50,
        n_supporting_variant_sequences=1,
        n_supporting_reference_transcripts=2,
        n_total_variant_sequences=3,
        n_total_variant_reads=100,
        n_total_reference_transcripts=5)

    protseq_most_reference_transcripts = make_dummy_protein_sequence(
        n_supporting_variant_reads=40,
        n_supporting_variant_sequences=1,
        n_supporting_reference_transcripts=3,
        n_total_variant_sequences=3,
        n_total_variant_reads=100,
        n_total_reference_transcripts=5)

    protseq_fewest_reads_or_transcripts = make_dummy_protein_sequence(
        n_supporting_variant_reads=10,
        n_supporting_variant_sequences=1,
        n_supporting_reference_transcripts=1,
        n_total_variant_sequences=3,
        n_total_variant_reads=100,
        n_total_reference_transcripts=5)
    unsorted_protein_sequences = [
        protseq_fewest_reads_or_transcripts,
        protseq_most_reads,
        protseq_most_reference_transcripts
    ]
    expected_order = [
        protseq_most_reads,
        protseq_most_reference_transcripts,
        protseq_fewest_reads_or_transcripts,
    ]
    eq_(sort_protein_sequences(unsorted_protein_sequences), expected_order)


def test_variants_to_protein_sequences_dataframe_one_sequence_per_variant():
    expressed_variants = load_vcf("data/b16.f10/b16.expressed.vcf")
    not_expressed_variants = load_vcf("data/b16.f10/b16.not-expressed.vcf")

    combined_variants = VariantCollection(
        list(expressed_variants) + list(not_expressed_variants))

    samfile = load_bam("data/b16.f10/b16.combined.sorted.bam")
    df = variants_to_protein_sequences_dataframe(
        combined_variants,
        samfile,
        min_mapping_quality=0,
        max_protein_sequences_per_variant=1)
    print(df)
    eq_(
        len(df),
        len(expressed_variants),
        "Expected %d/%d entries to have RNA support, got %d" % (
            len(expressed_variants),
            len(combined_variants),
            len(df),))

def test_variants_to_protein_sequences_dataframe_filtered_all_reads_by_mapping_quality():
    # since the B16 BAM has all MAPQ=255 values then all the reads should get dropped
    # if we set the minimum quality to 256
    variants = load_vcf("data/b16.f10/b16.vcf")
    samfile = load_bam("data/b16.f10/b16.combined.sorted.bam")
    df = variants_to_protein_sequences_dataframe(
        variants,
        samfile,
        min_mapping_quality=256,
        max_protein_sequences_per_variant=1)
    print(df)
    eq_(
        len(df),
        0,
        "Expected 0 entries, got %d" % (len(df),))

def test_variants_to_protein_sequences_dataframe_protein_sequence_length():
    variants = load_vcf("data/b16.f10/b16.expressed.vcf")
    samfile = load_bam("data/b16.f10/b16.combined.sorted.bam")

    for desired_length in range(9, 20, 3):
        df = variants_to_protein_sequences_dataframe(
            variants,
            samfile,
            max_protein_sequences_per_variant=1,
            protein_sequence_length=desired_length)
        eq_(
            len(df),
            len(variants),
            "Expected %d entries for protein_sequnece_length=%d, got %d results" % (
                len(variants),
                desired_length,
                len(df)))
        protein_sequences = df["amino_acids"]
        print(protein_sequences)
        protien_sequence_lengths = protein_sequences.str.len()
        assert (protien_sequence_lengths == desired_length).all(), protien_sequence_lengths

if __name__ == "__main__":
    test_sort_protein_sequences()
    test_variants_to_protein_sequences_dataframe_one_sequence_per_variant()
    test_variants_to_protein_sequences_dataframe_filtered_all_reads_by_mapping_quality()
    test_variants_to_protein_sequences_dataframe_protein_sequence_length()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from nose.tools import eq_
from varcode import Variant
from isovar.variant_sequence import variant_reads_to_sequences
from isovar.variant_read import gather_reads_for_single_variant

from testing_helpers import load_bam

def test_sequence_counts_snv():
    samfile = load_bam("data/cancer-wgs-primary.chr12.bam")
    chromosome = "chr12"
    base1_location = 65857041
    ref = "G"
    alt = "C"
    variant = Variant(chromosome, base1_location, ref, alt)

    variant_reads = gather_reads_for_single_variant(
        samfile=samfile,
        chromosome=chromosome,
        variant=variant)

    variant_sequences = variant_reads_to_sequences(
        variant_reads,
        max_nucleotides_before_variant=30,
        max_nucleotides_after_variant=30)
    assert len(variant_sequences) == 1
    for variant_sequence in variant_sequences:
        print(variant_sequence)
        eq_(variant_sequence.alt, alt)
        eq_(len(variant_sequence.prefix), 30)
        eq_(len(variant_sequence.suffix), 30)
        eq_(
            variant_sequence.prefix + variant_sequence.alt + variant_sequence.suffix,
            variant_sequence.full_sequence)

if __name__ == "__main__":
    test_sequence_counts_snv()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helper functions which it easier to write shell scripts
"""

from __future__ import print_function, division, absolute_import
import varcode
from pysam import AlignmentFile

from .protein_sequence import variants_to_protein_sequences_dataframe

def variants_to_protein_sequences_dataframe_from_args(args):
    variants = varcode.load_vcf(args.vcf, genome=args.genome)
    samfile = AlignmentFile(args.bam)
    return variants_to_protein_sequences_dataframe(
        variants=variants,
        samfile=samfile,
        protein_sequence_length=args.protein_sequence_length,
        min_reads_supporting_rna_sequence=args.min_reads,
        min_transcript_prefix_length=args.min_transcript_prefix_length,
        max_transcript_mismatches=args.max_reference_transcript_mismatches,
        max_protein_sequences_per_variant=args.max_protein_sequences_per_variant,
        min_mapping_quality=args.min_mapping_quality)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module combines variant cDNA sequences collected from a BAM file with
the reading frames of annotated reference transcripts to create candidate
translations.
"""

from __future__ import print_function, division, absolute_import
from collections import namedtuple
import math
import logging

from skbio import DNA

from .reference_context import reference_contexts_for_variant
from .variant_sequence import variant_sequences_generator
from .default_parameters import (
    MIN_TRANSCRIPT_PREFIX_LENGTH,
    MAX_REFERENCE_TRANSCRIPT_MISMATCHES,
    PROTEIN_SEQUENCE_LENGTH,
    MIN_READS_SUPPORTING_VARIANT_CDNA_SEQUENCE,
    MIN_READ_MAPPING_QUALITY,
)
from .dataframe_builder import DataFrameBuilder

# When multiple distinct RNA sequence contexts and/or reference transcripts
# give us the same translation then we group them into a single object
# which summarizes the supporting read count and reference transcript ORFs
# for a unique protein sequence.

################################
#
# VariantSequenceInReadingFrame
# ------------------------------
#
# Combines a VariantSequence with the reading frame implied by a
# ReferenceContext, reverse complementing if necessary and finding the
# offset to the first complete codon in the cDNA sequence.
#
#################################


VariantSequenceInReadingFrame = namedtuple(
    "VariantSequenceInReadingFrame",
    (
        # since the reference context and variant sequence may have
        # different numbers of nucleotides before the variant, the cDNA prefix
        # gets truncated to the shortest length. To avoid having to recompute
        # that sequence again, let's just cache the full cDNA sequence we used
        # for translation here, along with an interval indicating which
        # nucleotides are from the variant of interest
        "cdna_sequence",
        "offset_to_first_complete_codon",
        "variant_cdna_interval_start",
        "variant_cdna_interval_end",
        "reference_cdna_sequence_before_variant",
        "number_mismatches",
    ))


##########################
#
# TranslationKey
# --------------
#
# Fields related to a translated protein sequence which should be used to
# combine multiple equivalent sequences.
#
##########################


# pulling out the tuple of fields so that we can reuse them in the
# ProteinSequence namedtuple, as well as Translation
TranslationKey = namedtuple(
    "TranslationKey", (
        # translated sequence of a variant sequence in the ORF established
        # by a reference context
        "amino_acids",
        # half-open interval coordinates for variant amino acids
        # in the translated sequence
        "variant_aa_interval_start",
        "variant_aa_interval_end",
        # did the amino acid sequence end due to a stop codon or did we
        # just run out of sequence context around the variant?
        "ends_with_stop_codon",
        # was the variant a frameshift relative to the reference sequence?
        "frameshift"))


##########################
#
# Translation
# -----------
#
# Translated amino acid sequence of a VariantSequenceInReadingFrame for a
# particular ReferenceContext and VariantSequence.
#
##########################


Translation = namedtuple(
    "Translation",
    VariantSequenceInReadingFrame._fields + TranslationKey._fields + (
        "variant_sequence",
        "reference_context",
        "variant_sequence_in_reading_frame"))


def trim_sequences(variant_sequence, reference_context):
    """
    A VariantSequence and ReferenceContext may contain a different number of
    nucleotides before the variant locus. Furthermore, the VariantSequence is
    always expressed in terms of the positive strand against which it aligned,
    but reference transcripts may have sequences from the negative strand of the
    genome. Take the reverse complement of the VariantSequence if the
    ReferenceContext is from negative strand transcripts and trim either
    sequence to ensure that the prefixes are of the same length.

    Parameters
    ----------
    variant_sequence : VariantSequence

    reference_context : ReferenceContext

    Returns a tuple with the following fields:
        1) cDNA prefix of variant sequence, trimmed to be same length as the
           reference prefix. If the reference context was on the negative
           strand then this is the trimmed sequence *after* the variant from
           the genomic DNA sequence.
        2) cDNA sequence of the variant nucleotides, in reverse complement if
           the reference context is on the negative strand.
        3) cDNA sequence of the nucleotides after the variant nucleotides. If
           the reference context is on the negative strand then this sequence
           is the reverse complement of the original prefix sequence.
        4) Reference sequence before the variant locus, trimmed to be the
           same length as the variant prefix.
        5) Number of nucleotides trimmed from the reference sequence, used
           later for adjustint offset to first complete codon.
    """
    cdna_prefix = DNA(variant_sequence.prefix)
    cdna_alt = DNA(variant_sequence.alt)
    cdna_suffix = DNA(variant_sequence.suffix)

    # if the transcript is on the reverse strand then we have to
    # take the sequence PREFIX|VARIANT|SUFFIX
    # and take the complement of XIFFUS|TNAIRAV|XIFERP
    if reference_context.strand == "-":
        # notice that we are setting the *prefix* to be reverse complement
        # of the *suffix* and vice versa
        cdna_prefix, cdna_alt, cdna_suffix = (
            cdna_suffix.reverse_complement(),
            cdna_alt.reverse_complement(),
            cdna_prefix.reverse_complement()
        )

    reference_sequence_before_variant = reference_context.sequence_before_variant_locus

    # trim the reference sequence and the RNA-derived sequence to the same length
    if len(reference_sequence_before_variant) > len(cdna_prefix):
        n_trimmed_from_reference = len(reference_sequence_before_variant) - len(cdna_prefix)
        n_trimmed_from_variant = 0
    elif len(reference_sequence_before_variant) < len(cdna_prefix):
        n_trimmed_from_variant = len(cdna_prefix) - len(reference_sequence_before_variant)
        n_trimmed_from_reference = 0
    else:
        n_trimmed_from_variant = 0
        n_trimmed_from_reference = 0

    reference_sequence_before_variant = reference_sequence_before_variant[
        n_trimmed_from_reference:]

    cdna_prefix = cdna_prefix[n_trimmed_from_variant:]

    return (
        cdna_prefix,
        cdna_alt,
        cdna_suffix,
        reference_sequence_before_variant,
        n_trimmed_from_reference
    )


def count_mismatches(reference_prefix, cdna_prefix):
    """
    Computes the number of mismatching nucleotides between two cDNA sequences.

    Parameters
    ----------
    reference_prefix : str or skbio.DNA
        cDNA sequence of a reference transcript before a variant locus

    cdna_prefix : str or skbio.DNA
        cDNA sequence detected from RNAseq before a variant locus
    """
    if len(reference_prefix) != len(cdna_prefix):
        raise ValueError(
            "Expected reference prefix '%s' to be same length as %s" % (
                reference_prefix, cdna_prefix))
    # converting to str before zip because skbio.DNA is weird and sometimes
    # returns elements which don't compare equally despite being the same
    # nucleotide
    return sum(xi != yi for (xi, yi) in zip(
        str(reference_prefix), str(cdna_prefix)))


def compute_offset_to_first_complete_codon(
        offset_to_first_complete_reference_codon,
        n_trimmed_from_reference_sequence):
    """
    Once we've aligned the variant sequence to the ReferenceContext, we need
    to transfer reading frame from the reference transcripts to the variant
    sequences.

    Parameters
    ----------
    offset_to_first_complete_reference_codon : ReferenceContext

    n_trimmed_from_reference_sequence : int

    Returns an offset into the variant sequence that starts from a complete
    codon.
    """
    if n_trimmed_from_reference_sequence <= offset_to_first_complete_reference_codon:
        return (
            offset_to_first_complete_reference_codon -
            n_trimmed_from_reference_sequence)
    else:
        n_nucleotides_trimmed_after_first_codon = (
            n_trimmed_from_reference_sequence -
            offset_to_first_complete_reference_codon)
        frame = n_nucleotides_trimmed_after_first_codon % 3
        return (3 - frame) % 3

def determine_reading_frame_for_variant_sequence(variant_sequence, reference_context):
    """
    Parameters
    ----------
    variant_sequence : VariantSequence

    reference_context : ReferenceContext

    Returns a VariantSequenceInReadingFrame object
    """
    cdna_prefix, cdna_alt, cdna_suffix, reference_prefix, n_trimmed_from_reference = \
        trim_sequences(variant_sequence, reference_context)

    n_mismatch_before_variant = count_mismatches(reference_prefix, cdna_prefix)

    # ReferenceContext carries with an offset to the first complete codon
    # in the reference sequence. This may need to be adjusted if the reference
    # sequence is longer than the variant sequence (and thus needs to be trimmed)
    offset_to_first_complete_codon = compute_offset_to_first_complete_codon(
        offset_to_first_complete_reference_codon=reference_context.offset_to_first_complete_codon,
        n_trimmed_from_reference_sequence=n_trimmed_from_reference)

    cdna_sequence = DNA.concat([cdna_prefix, cdna_alt, cdna_suffix])
    variant_interval_start = len(cdna_prefix) + 1
    variant_interval_end = variant_interval_start + len(cdna_alt)

    return VariantSequenceInReadingFrame(
        cdna_sequence=cdna_sequence,
        offset_to_first_complete_codon=offset_to_first_complete_codon,
        variant_cdna_interval_start=variant_interval_start,
        variant_cdna_interval_end=variant_interval_end,
        reference_cdna_sequence_before_variant=reference_prefix,
        number_mismatches=n_mismatch_before_variant)

def find_mutant_amino_acid_interval(
        cdna_sequence,
        cdna_first_codon_offset,
        cdna_variant_start_offset,
        cdna_variant_end_offset,
        n_ref,
        n_amino_acids):
    """
    Parameters
    ----------
    cdna_sequence : skbio.DNA or str
        cDNA sequence found in RNAseq data

    cdna_first_codon_offset : int
        Offset into cDNA sequence to first complete codon, lets us skip
        past UTR region and incomplete codons.

    cdna_variant_start_offset : int
        Interbase start offset into cDNA sequence for selecting mutant
        nucleotides.

    cdna_variant_end_offset : int
        Interbase end offset into cDNA sequence for selecting mutant
        nucleotides.

    n_ref : int
        Number of reference nucleotides

    n_amino_acids : int
        Number of translated amino acids

    Returns tuple with three fields:
        1) Start offset for interval of mutant amino acids in translated sequence
        2) End offset for interval of mutant amino acids in translated sequence
        3) Boolean flag indicating whether the variant was a frameshift.
    """
    cdna_alt_nucleotides = cdna_sequence[
        cdna_variant_start_offset:cdna_variant_end_offset]

    n_alt = len(cdna_alt_nucleotides)

    # sequence of nucleotides before the variant starting from the first codon
    cdna_coding_prefix = cdna_sequence[cdna_first_codon_offset:cdna_variant_start_offset]

    # rounding down since a change in the middle of a codon should count
    # toward the variant codons
    n_coding_nucleotides_before_variant = len(cdna_coding_prefix)
    n_complete_prefix_codons = cdna_variant_start_offset // 3

    frame_of_variant_nucleotides = n_coding_nucleotides_before_variant % 3
    frameshift = abs(n_ref - n_alt) % 3 != 0
    indel = n_ref != n_alt

    variant_aa_interval_start = n_complete_prefix_codons + 1
    if frameshift:
        # if mutation is a frame shift then every amino acid from the
        # first affected codon to the stop is considered mutant
        #
        # TODO: what if the first k amino acids are synonymous with the
        # reference sequence?
        variant_aa_interval_end = n_amino_acids
    else:
        n_alt_codons = int(math.ceil(n_alt / 3.0))
        if indel:
            # We need to adjust the number of affected codons by whether the
            # variant is aligned with codon boundaries, since in-frame indels
            # may still be split across multiple codons.
            #
            # Example of in-frame deletion of 3 nucleotides which leaves
            # 0 variant codons in the sequence (interval = 1:1)
            #   ref = CCC|AAA|GGG|TTT
            #   alt = CCC|GGG|TTT
            #
            # Example of in-frame deletion of 3 nucleotides which leaves
            # 1 variant codon in the sequence (interval = 1:2)
            #   ref = CCC|AAA|GGG|TTT
            #   alt = CCC|AGG|TTT
            #
            # Example of in-frame insertion of 3 nucleotides which
            # yields two variant codons:
            #   ref = CCC|AAA|GGG|TTT
            #   alt = CTT|TCC|AAA|GGG|TTT
            extra_affected_codon = int(frame_of_variant_nucleotides != 0)
            variant_aa_interval_end = (
                variant_aa_interval_start + n_alt_codons + extra_affected_codon)
        else:
            # if the variant is a simple substitution then it only affects
            # as many codons as are in the alternate sequence
            variant_aa_interval_end = variant_aa_interval_start + n_alt_codons
    return variant_aa_interval_start, variant_aa_interval_end, frameshift

def translate_cdna(cdna_sequence_from_first_codon):
    """
    Given a cDNA sequence which is aligned to a reading frame, returns
    the translated protein sequence and a boolean flag indicating whether
    the translated sequence ended on a stop codon (or just ran out of codons).

    Parameters
    ----------
    cdna_sequence_from_first_codon : skbio.DNA
        cDNA sequence which is expected to start and end on complete codons.
    """
    # once we drop some of the prefix nucleotides, we should be in a reading frame
    # which allows us to translate this protein
    variant_amino_acids = cdna_sequence_from_first_codon.translate()

    # by default, the translate method emits "*" for
    # the stop codons TAA, TAG and TGG
    ends_with_stop_codon = "*" in variant_amino_acids

    if ends_with_stop_codon:
        # truncate the amino acid sequence at the stop codon
        stop_codon_index = variant_amino_acids.index("*")
        variant_amino_acids = variant_amino_acids[:stop_codon_index]

    return variant_amino_acids, ends_with_stop_codon


def translate_variant_sequence(
        variant_sequence,
        reference_context,
        max_transcript_mismatches,
        protein_sequence_length=None):
    """
    Attempt to translate a single VariantSequence using the reading frame
    from a single ReferenceContext.

    Parameters
    ----------
    variant_sequence : VariantSequence

    reference_context : ReferenceContext

    max_transcript_mismatches : int
        Don't use the reading frame from a context where the cDNA variant
        sequences disagrees at more than this number of positions before the
        variant nucleotides.

    protein_sequence_length : int, optional
        Truncate protein to be at most this long

    Returns either a ProteinSequence object or None if the number of
    mismatches between the RNA and reference transcript sequences exceeds the
    given threshold.
    """

    variant_sequence_in_reading_frame = determine_reading_frame_for_variant_sequence(
        variant_sequence=variant_sequence,
        reference_context=reference_context)

    n_mismatch_before_variant = variant_sequence_in_reading_frame.number_mismatches

    if n_mismatch_before_variant > max_transcript_mismatches:
        logging.info(
            "Skipping reference context %s for %s, too many mismatching bases (%d)",
            reference_context,
            variant_sequence,
            n_mismatch_before_variant)
        return None

    cdna_sequence = variant_sequence_in_reading_frame.cdna_sequence
    cdna_codon_offset = variant_sequence_in_reading_frame.offset_to_first_complete_codon
    cdna_sequence_from_first_codon = cdna_sequence[cdna_codon_offset:]

    variant_amino_acids, ends_with_stop_codon = translate_cdna(cdna_sequence_from_first_codon)

    # get the offsets into the cDNA sequence which pick out the variant nucleotides
    cdna_variant_start_offset = variant_sequence_in_reading_frame.variant_cdna_interval_start
    cdna_variant_end_offset = variant_sequence_in_reading_frame.variant_cdna_interval_end

    variant_aa_interval_start, variant_aa_interval_end, frameshift = \
        find_mutant_amino_acid_interval(
            cdna_sequence=cdna_sequence,
            cdna_first_codon_offset=cdna_codon_offset,
            cdna_variant_start_offset=cdna_variant_start_offset,
            cdna_variant_end_offset=cdna_variant_end_offset,
            n_ref=len(reference_context.sequence_at_variant_locus),
            n_amino_acids=len(variant_amino_acids))

    if protein_sequence_length and len(variant_amino_acids) > protein_sequence_length:
        if protein_sequence_length <= variant_aa_interval_start:
            logging.warn(
                ("Truncating amino acid sequence %s from variant sequence %s "
                 "to only %d elements loses all variant residues") % (
                    variant_amino_acids,
                    variant_sequence,
                    protein_sequence_length))
            return None
        # if the protein is too long then short it, which implies we're no longer
        # stopping due to a stop codon and that the variant amino acids might
        # need a new stop index
        variant_amino_acids = variant_amino_acids[:protein_sequence_length]
        variant_aa_interval_end = min(variant_aa_interval_end, protein_sequence_length)
        ends_with_stop_codon = False

    reference_sequence_before_variant = (
        variant_sequence_in_reading_frame.reference_cdna_sequence_before_variant)
    # converting sequence objects to str since skbio.DNA objects aren't really
    # useful outside of the translation module & debugging
    return Translation(
        cdna_sequence=str(cdna_sequence),
        offset_to_first_complete_codon=cdna_codon_offset,
        variant_cdna_interval_start=cdna_variant_start_offset,
        variant_cdna_interval_end=cdna_variant_end_offset,
        reference_cdna_sequence_before_variant=reference_sequence_before_variant,
        number_mismatches=variant_sequence_in_reading_frame.number_mismatches,
        amino_acids=str(variant_amino_acids),
        frameshift=frameshift,
        ends_with_stop_codon=ends_with_stop_codon,
        variant_aa_interval_start=variant_aa_interval_start,
        variant_aa_interval_end=variant_aa_interval_end,
        variant_sequence=variant_sequence,
        reference_context=reference_context,
        variant_sequence_in_reading_frame=variant_sequence_in_reading_frame)


def translation_generator(
        variant_sequences,
        reference_contexts,
        max_transcript_mismatches,
        protein_sequence_length=None):
    """
    Given all detected VariantSequence objects for a particular variant
    and all the ReferenceContext objects for that locus, translate
    multiple protein sequences, up to the number specified by the argument
    max_protein_sequences_per_variant.

    Parameters
    ----------
    variant_sequences : list of VariantSequence objects
        Variant sequences overlapping a single original variant

    reference_contexts : list of ReferenceContext objects
        Reference sequence contexts from the same variant as the variant_sequences

    max_transcript_mismatches : int

    protein_sequence_length : int, optional
        Truncate protein to be at most this long

    Yields a sequence of Translation objects.
    """
    for reference_context in reference_contexts:
        for variant_sequence in variant_sequences:
            translation = translate_variant_sequence(
                variant_sequence=variant_sequence,
                reference_context=reference_context,
                max_transcript_mismatches=max_transcript_mismatches,
                protein_sequence_length=protein_sequence_length)
            if translation is not None:
                yield translation

def translate_variants(
        variants,
        samfile,
        transcript_id_whitelist=None,
        protein_sequence_length=PROTEIN_SEQUENCE_LENGTH,
        min_reads_supporting_rna_sequence=MIN_READS_SUPPORTING_VARIANT_CDNA_SEQUENCE,
        min_transcript_prefix_length=MIN_TRANSCRIPT_PREFIX_LENGTH,
        max_transcript_mismatches=MAX_REFERENCE_TRANSCRIPT_MISMATCHES,
        min_mapping_quality=MIN_READ_MAPPING_QUALITY):
    """
    Translates each coding variant in a collection to one or more protein
    fragment sequences (if the variant is not filtered and its spanning RNA
    sequences can be given a reading frame).

    Parameters
    ----------
    variants : varcode.VariantCollection

    samfile : pysam.AlignmentFile

    transcript_id_whitelist : set, optional
        If given, expected to be a set of transcript IDs which we should use
        for determining the reading frame around a variant. If omitted, then
        try to use all overlapping reference transcripts.

    protein_sequence_length : int
        Try to translate protein sequences of this length, though sometimes
        we'll have to return something shorter (depending on the RNAseq data,
        and presence of stop codons).

    min_reads_supporting_rna_sequence : int
        Drop variant sequences supported by fewer than this number of reads.

    min_transcript_prefix_length : int
        Minimum number of bases we need to try matching between the reference
        context and variant sequence.

    max_transcript_mismatches : int
        Don't try to determine the reading frame for a transcript if more
        than this number of bases differ.

     min_mapping_quality : int
        Minimum MAPQ value before a read gets ignored

    Yields pairs of a Variant and a generator of all its candidate
    Translation objects.
    """

    # Adding an extra codon to the desired RNA sequence length in case we
    # need to clip nucleotides at the start/end of the sequence
    rna_sequence_length = (protein_sequence_length + 1) * 3

    for variant, variant_sequences in variant_sequences_generator(
            variants=variants,
            samfile=samfile,
            sequence_length=rna_sequence_length,
            min_reads=min_reads_supporting_rna_sequence,
            min_mapping_quality=min_mapping_quality):

        if len(variant_sequences) == 0:
            logging.info(
                "Skipping variant %s, no cDNA sequences detected" % (
                    variant,))
            yield (variant, [])
            continue

        # try translating the variant sequences from the same set of
        # ReferenceContext objects, which requires using the longest
        # context_size to be compatible with all of the sequences. Some
        # sequences maybe have fewer nucleotides than this before the variant
        # and will thus have to be trimmed.
        context_size = max(
            len(variant_sequence.prefix)
            for variant_sequence in variant_sequences)

        if context_size < min_transcript_prefix_length:
            logging.info(
                "Skipping variant %s, none of the cDNA sequences have sufficient context" % (
                    variant,))
            yield (variant, [])
            continue

        reference_contexts = reference_contexts_for_variant(
            variant,
            context_size=context_size,
            transcript_id_whitelist=transcript_id_whitelist)

        translations = translation_generator(
            variant_sequences=variant_sequences,
            reference_contexts=reference_contexts,
            max_transcript_mismatches=max_transcript_mismatches,
            protein_sequence_length=protein_sequence_length)

        yield variant, translations


def translate_variants_dataframe(*args, **kwargs):
    """
    Given a collection of variants and a SAM/BAM file of overlapping reads,
    returns a DataFrame of translated protein fragments with columns
    for each field of a Translation object (and chr/pos/ref/alt per variant).

    Takes the same parameters as translate_variants.
    """

    df_builder = DataFrameBuilder(
        Translation,
        # exlude fields which are structured objects
        exclude=[
            "variant_sequence",
            "reference_context",
            "variant_sequence_in_reading_frame"])

    for (variant, translations) in translate_variants(*args, **kwargs):
        for translation in translations:
            df_builder.add(variant, translation)
    return df_builder.to_dataframe()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from isovar.variant_read import gather_reads_for_single_variant
from isovar.assembly import assemble_transcript_fragments
from pyensembl import ensembl_grch38
from varcode import Variant
from nose.tools import eq_

from testing_helpers import load_bam

def test_assemble_transcript_fragments_snv():
    samfile = load_bam("data/cancer-wgs-primary.chr12.bam")
    chromosome = "chr12"
    base1_location = 65857041
    ref = "G"
    alt = "C"
    variant = Variant(
        contig=chromosome,
        start=base1_location,
        ref=ref,
        alt=alt,
        ensembl=ensembl_grch38)
    variant_reads = gather_reads_for_single_variant(
        samfile=samfile,
        chromosome=chromosome,
        variant=variant)

    longer_sequences = assemble_transcript_fragments(
        variant_reads,
        min_overlap_size=30)

    assert len(longer_sequences) > 0

    for (p, v, s), c in longer_sequences:
        print("%s%s%s weight=%d length=%d" % (
            p,
            v,
            s,
            c,
            len(p + v + s)))
        eq_(v, alt)

if __name__ == "__main__":
    test_assemble_transcript_fragments_snv()



#!/usr/bin/env python

# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Translate non-synonymous coding variants into mutant protein sequences using an
RNAseq BAM from the same sample. Combine synonymous translations and assign
a read count to each protein sequence.
"""

from __future__ import print_function, division, absolute_import
import logging
import argparse

from isovar.args import (
    add_somatic_vcf_args,
    add_rna_args,
    add_reference_context_args,
    add_protein_sequence_args
)
from isovar.cli_helpers import variants_to_protein_sequences_dataframe_from_args


parser = argparse.ArgumentParser()
parser = add_somatic_vcf_args(parser)
parser = add_rna_args(parser)
parser = add_reference_context_args(parser)
parser = add_protein_sequence_args(parser)
parser.add_argument(
    "--output",
    default="isovar-translate-variants-results.csv",
    help="Name of CSV file which contains predicted sequences")

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    args = parser.parse_args()
    print(args)
    df = variants_to_protein_sequences_dataframe_from_args(args)
    print(df)
    df.to_csv(args.output)



from argparse import ArgumentParser
from isovar.args import (
    add_somatic_vcf_args,
    add_rna_args,
    add_reference_context_args,
    add_protein_sequence_args,
)
from isovar.cli_helpers import variants_to_protein_sequences_dataframe_from_args

from testing_helpers import data_path

def test_variants_to_protein_sequences_dataframe_from_args():
    vcf_path = data_path("data/b16.f10/b16.vcf")
    bam_path = data_path("data/b16.f10/b16.combined.bam")

    parser = \
        add_protein_sequence_args(
            add_reference_context_args(
                add_rna_args(
                    add_somatic_vcf_args(ArgumentParser()))))
    args = parser.parse_args([
        "--vcf", vcf_path,
        "--bam", bam_path,
        "--genome", "mm10",
    ])
    df = variants_to_protein_sequences_dataframe_from_args(args)
    assert len(df) > 0



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import
import logging

from collections import defaultdict

nucleotides = ["A", "C", "T", "G"]
nucleotide_to_index = {c: i for (i, c) in enumerate(nucleotides)}
index_to_nucleotide = {i: c for (i, c) in enumerate(nucleotides)}

def create_logger(
        name="root",
        level=logging.DEBUG,
        format_string="%(levelname)s: %(message)s (%(filename)s:%(lineno)s - %(funcName)s)"):
    logger = logging.getLogger(name)
    handler = logging.StreamHandler()
    formatter = logging.Formatter(format_string)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(level)
    return logger

logger = create_logger()

def group_unique_sequences(
        variant_reads,
        max_prefix_size=None,
        max_suffix_size=None):
    """
    Given a list of VariantRead objects, extracts all unique
    (prefix, suffix) sequence pairs and associate each with a list
    of read names that contained that sequence.
    """
    groups = defaultdict(set)
    for r in variant_reads:
        prefix = r.prefix
        suffix = r.suffix
        if max_prefix_size and len(prefix) > max_prefix_size:
            prefix = prefix[-max_prefix_size:]
        if max_suffix_size and len(suffix) > max_suffix_size:
            suffix = suffix[:max_suffix_size]
        key = (prefix, suffix)
        groups[key].add(r.name)
    return groups

def count_unique_sequences(
        variant_reads,
        max_prefix_size=None,
        max_suffix_size=None):
    """
    Given a list of VariantRead objects, extracts all unique
    (prefix, suffix) sequence pairs and associate each with the number
    of reads that contain that sequence.
    """
    groups = group_unique_sequences(
        variant_reads,
        max_prefix_size=max_prefix_size,
        max_suffix_size=max_suffix_size)
    return {
        seq_pair: len(read_names)
        for (seq_pair, read_names) in groups.items()
    }


def get_variant_nucleotides(variant_reads):
    if len(variant_reads) > 0:
        variant_seq = variant_reads[0].alt
        if not all(r.alt == variant_seq for r in variant_reads):
            raise ValueError(
                ("Cannot call `get_variant_nucleotides` on a collection"
                 " of VariantRead objects spanning multiple variants"))
        return variant_seq
    else:
        raise ValueError("Expected len(variant_reads) > 0")


def make_prefix_suffix_pairs(variant_reads):
    variant_seq = get_variant_nucleotides(variant_reads)
    pairs = [(r.prefix, r.suffix) for r in variant_reads]
    return variant_seq, pairs


def list_to_string(list_of_anything, sep=";"):
    """
    Helper function used for building the fields of a printable dataframe
    """
    return sep.join(str(x) for x in list_of_anything)



import isovar
import isovar.variant_read
from isovar.variant_read import (
    variant_read_from_single_read_at_locus,
    VariantRead,
)
from isovar.read_at_locus import ReadAtLocus
from nose.tools import eq_

def make_read_at_locus(prefix, alt, suffix, base_quality=30, name="dummy"):
    dummy_sequence = prefix + alt + suffix
    return ReadAtLocus(
        name="dummy",
        sequence=dummy_sequence,
        reference_positions=list(range(1, len(dummy_sequence) + 1)),
        quality_scores=[base_quality] * len(dummy_sequence),
        base0_read_position_before_variant=len(prefix) - 1,
        base0_read_position_after_variant=len(prefix) + len(alt),
    )

def test_variant_read_from_single_read_at_locus_trim_N_nucleotides():
    read_at_locus =make_read_at_locus(prefix="NCCN", alt="A", suffix="TNNA")
    variant_read = variant_read_from_single_read_at_locus(
        read_at_locus, ref="T", alt="A")
    print(variant_read)
    expected = VariantRead(prefix="", alt="A", suffix="T", name="dummy")
    eq_(variant_read, expected)

if __name__ == "__main__":
    test_variant_read_from_single_read_at_locus_trim_N_nucleotides()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Since multiple variant sequences can translate to the same amino acid sequence,
this module aggregates equivalent Translation objects into a single
ProteinSequence.
"""

from __future__ import print_function, division, absolute_import
from collections import namedtuple, defaultdict

from .default_parameters import (
    MIN_TRANSCRIPT_PREFIX_LENGTH,
    MAX_REFERENCE_TRANSCRIPT_MISMATCHES,
    PROTEIN_SEQUENCE_LENGTH,
    MAX_PROTEIN_SEQUENCES_PER_VARIANT,
    MIN_READS_SUPPORTING_VARIANT_CDNA_SEQUENCE,
    MIN_READ_MAPPING_QUALITY,
)
from .dataframe_builder import DataFrameBuilder
from .translation import translate_variants, TranslationKey

##########################
#
# ProteinSequence
# ---------------
#
# Translated amino acid sequence aggregated across possibly multiple
# VariantSequence and ReferenceContext objects (e.g. imagine two distinct
# sequences which contain synonymous codons).
#
# This is the final result of the isovar variant->expressed peptide pipeline.
#
##########################

ProteinSequence = namedtuple(
    "ProteinSequence",
    TranslationKey._fields + (
        # list of all the Translation objects which support this distinct
        # amino acid sequence
        "translations",
        # number of unique read names from all the VariantSequence objects
        # from each translation
        "supporting_variant_reads",
        # total number of reads at the locus which contained the variant
        # nucleotides, even if they supported other phased sequences
        "total_variant_reads",
        # how many reference transcripts were used to establish the
        # reading frame for this protein sequence
        "supporting_transcripts",
        # how many reference transcripts overlap the variant locus?
        "total_transcripts",
        # name of gene of the reference transcripts used in Translation
        # objects
        "gene",
    ))


def to_translation_key(x):
    """
    If a namedtuple has a superset of the fields of a TranslationKey,
    project it down to create a TranslationKey object.
    """
    values_dict = {
        name: getattr(x, name)
        for name in TranslationKey._fields
    }
    return TranslationKey(**values_dict)

def translation_key_to_protein_sequence(translation_key, **extra_kwargs):
    """
    Create a ProteinSequence object from a TranslationKey, and extra fields
    which must be supplied in extra_kwargs.
    """
    values_dict = {}
    for name in TranslationKey._fields:
        values_dict[name] = getattr(translation_key, name)
    values_dict.update(extra_kwargs)
    return ProteinSequence(**values_dict)

def summarize_translations(translations):
    """
    Summarize a collection of Translation objects into three values:
        1) Set of unique read names supporting underlying variant sequences
        2) Set of unique transcript names for establishing reading frames of the
           translations.
        3) Set of unique gene names for all transcripts used by translations.
    """
    read_names = set([])
    gene_names = set([])
    transcript_ids = set([])
    for translation in translations:
        for read_name in translation.variant_sequence.read_names:
            read_names.add(read_name)
        for transcript in translation.reference_context.transcripts:
            transcript_ids.add(transcript.id)
            gene_names.add(transcript.gene.name)
    return read_names, transcript_ids, gene_names

def protein_sequence_sort_key(protein_sequence):
    """
    Sort protein sequences lexicographically by three criteria:
        - number of unique supporting reads
        - minimum mismatch versus a supporting reference transcript
        - number of supporting reference transcripts
    """
    return (
        len(protein_sequence.supporting_variant_reads),
        min(
            t.number_mismatches
            for t in protein_sequence.translations),
        len(protein_sequence.supporting_transcripts)
    )

def sort_protein_sequences(protein_sequences):
    """
    Sort protein sequences in decreasing order of priority
    """
    return list(
        sorted(protein_sequences, key=protein_sequence_sort_key, reverse=True))

def variants_to_protein_sequences(
        variants,
        samfile,
        transcript_id_whitelist=None,
        protein_sequence_length=PROTEIN_SEQUENCE_LENGTH,
        min_reads_supporting_rna_sequence=MIN_READS_SUPPORTING_VARIANT_CDNA_SEQUENCE,
        min_transcript_prefix_length=MIN_TRANSCRIPT_PREFIX_LENGTH,
        max_transcript_mismatches=MAX_REFERENCE_TRANSCRIPT_MISMATCHES,
        max_protein_sequences_per_variant=MAX_PROTEIN_SEQUENCES_PER_VARIANT,
        min_mapping_quality=MIN_READ_MAPPING_QUALITY):
    """"
    Translates each coding variant in a collection to one or more
    Translation objects, which are then aggregated into equivalent
    ProteinSequence objects.

    Parameters
    ----------
    variants : varcode.VariantCollection

    samfile : pysam.AlignmentFile

    transcript_id_whitelist : set, optional
        If given, expected to be a set of transcript IDs which we should use
        for determining the reading frame around a variant. If omitted, then
        try to use all overlapping reference transcripts.

    protein_sequence_length : int
        Try to translate protein sequences of this length, though sometimes
        we'll have to return something shorter (depending on the RNAseq data,
        and presence of stop codons).

    min_reads_supporting_rna_sequence : int
        Drop variant sequences supported by fewer than this number of reads.

    min_transcript_prefix_length : int
        Minimum number of bases we need to try matching between the reference
        context and variant sequence.

    max_transcript_mismatches : int
        Don't try to determine the reading frame for a transcript if more
        than this number of bases differ.

    max_protein_sequences_per_variant : int
        Number of protein sequences to return for each ProteinSequence

    min_mapping_quality : int
        Minimum MAPQ value before a read gets ignored

    Yields pairs of a Variant and a list of ProteinSequence objects
    """

    for (variant, translations) in translate_variants(
            variants=variants,
            samfile=samfile,
            transcript_id_whitelist=transcript_id_whitelist,
            protein_sequence_length=protein_sequence_length,
            min_reads_supporting_rna_sequence=min_reads_supporting_rna_sequence,
            min_transcript_prefix_length=min_transcript_prefix_length,
            max_transcript_mismatches=max_transcript_mismatches,
            min_mapping_quality=min_mapping_quality):

        # convert to list so we can traverse it twice, once to get TranslationKey
        # mappings for each Translation and a second time to collect all the
        # unique read names and transript IDs.
        translations = list(translations)

        # dictionary mapping TranslationKey object to a list of Translations
        # with the same protein sequence
        equivalent_translations_dict = defaultdict(list)

        for translation in translations:
            key = to_translation_key(translation)
            equivalent_translations_dict[key].append(translation)

        all_read_names, all_transcript_ids, _ = summarize_translations(translations)
        n_total_read_names = len(all_read_names)
        n_total_transcripts = len(all_transcript_ids)

        protein_sequences = []

        for (key, equivalent_translations) in equivalent_translations_dict.items():
            # get the variant read names, transcript IDs and gene names for
            # protein sequence we're about to construct
            group_read_names, group_transcript_ids, group_gene_names = \
                summarize_translations(equivalent_translations)

            protein_sequence = translation_key_to_protein_sequence(
                translation_key=key,
                translations=equivalent_translations,
                supporting_variant_reads=group_read_names,
                total_variant_reads=n_total_read_names,
                supporting_transcripts=group_transcript_ids,
                total_transcripts=n_total_transcripts,
                gene=list(group_gene_names))
            protein_sequences.append(protein_sequence)

        # sort protein sequences before returning the top results
        protein_sequences = sort_protein_sequences(protein_sequences)

        yield variant, protein_sequences[:max_protein_sequences_per_variant]

def variants_to_protein_sequences_dataframe(*args, **kwargs):
    """
    Given a collection of variants and a SAM/BAM file of overlapping reads,
    returns a DataFrame with a row for each protein sequence.

    Takes the same parameters as variants_to_protein_sequences.
    """
    df_builder = DataFrameBuilder(
        ProteinSequence,
        converters=dict(
            translations=len,
            supporting_variant_reads=len,
            supporting_transcripts=len,
            gene=lambda x: ";".join(x)))
    for (variant, protein_sequences) in variants_to_protein_sequences(*args, **kwargs):
        for protein_sequence in protein_sequences:
            df_builder.add(variant, protein_sequence)
    return df_builder.to_dataframe()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from .common import group_unique_sequences, get_variant_nucleotides


def sort_by_decreasing_prefix_length(x):
    return -len(x[0][0])

def sort_by_decreasing_suffix_length(x):
    return -len(x[0][1])

def sort_by_increasing_total_length(x):
    return (len(x[0][0]) + len(x[0][1]))

def sort_by_decreasing_total_length(x):
    return -(len(x[0][0]) + len(x[0][1]))

def greedy_merge(prefix_suffix_pair_groups, min_overlap_size=30):
    """
    Greedily merge overlapping sequences into longer sequences.

    Returns a dictionary mapping (prefix, suffix) sequence pairs to
    counts of how many reads support the new longer sequences.
    """
    merged = {}
    for (p1, s1), names1 in sorted(
            prefix_suffix_pair_groups.items(),
            key=sort_by_decreasing_prefix_length):
        len1 = len(p1) + len(s1)
        for (p2, s2), names2 in sorted(
                prefix_suffix_pair_groups.items(),
                key=sort_by_decreasing_suffix_length):
            if len(p2) > len(p1):
                continue
            elif len(s2) < len(s1):
                continue
            possible_overlap = len(p2) + len(s1)

            if possible_overlap < min_overlap_size:
                continue

            len2 = len(p2) + len(s2)
            new_length = len(p1) + len(s2)
            # if we're not extending the sequence length, then why bother?
            if new_length <= len1 or new_length <= len2:
                continue
            elif p1.endswith(p2) and s2.startswith(s1):
                # is the candidate sequence is a prefix of the accepted?
                # Example:
                # p1 s1 = XXXXXXXX ZZZZZZ
                # p2 s2 =       XX ZZZZZZZZZ
                # ...
                # then combine them into a longer sequence
                longer_pair = (p1, s2)
                combined_read_names = names1.union(names2)
                if longer_pair in merged:
                    merged[longer_pair] = merged[longer_pair].union(
                        combined_read_names)
                else:
                    merged[longer_pair] = combined_read_names
    return merged

def collapse_substrings(assembly_groups):
    """
    Combine shorter sequences which are fully contained in longer sequences
    """
    assert len(assembly_groups) > 0
    sorted_pairs = list(sorted(
        assembly_groups.items(),
        key=sort_by_decreasing_total_length))
    result_list = []
    for ((prefix_short, suffix_short), names_short) in sorted_pairs:
        found_superstring = False
        for (prefix_long, suffix_long), names_long in result_list:
            if prefix_long.endswith(prefix_short) and suffix_long.startswith(
                    suffix_short):
                found_superstring = True
                for name in names_short:
                    names_long.add(name)
        if not found_superstring:
            result_list.append(
                ((prefix_short, suffix_short), names_short.copy()))
    return {
        (p, s): names
        for ((p, s), names) in result_list
    }

def recursive_assembly(variant_reads, min_overlap_size=30, n_merge_iters=2):
    """
    Assembles longer sequences from reads centered on a variant by alternating
    between merging all pairs of overlapping sequences and collapsing
    shorter sequences onto every longer sequence which contains them.
    """
    assert len(variant_reads) > 0
    assembly_groups = group_unique_sequences(variant_reads)
    for i in range(n_merge_iters):
        previous_assembly_groups = {k: v for (k, v) in assembly_groups.items()}

        assembly_groups = greedy_merge(
            assembly_groups,
            min_overlap_size=min_overlap_size)

        if len(assembly_groups) == 0:
            return previous_assembly_groups
        assembly_groups = collapse_substrings(assembly_groups)
        if len(assembly_groups) == 1:
            return assembly_groups
    return assembly_groups

def final_sort_key(x):
    supporting_read_names = x[1]
    n_supporting_reads = len(supporting_read_names)
    prefix = x[0][0]
    suffix = x[0][1]
    combined_length = len(prefix) + len(suffix)
    return (n_supporting_reads, combined_length)

def assemble_transcript_fragments(
        variant_reads,
        min_overlap_size=30,
        n_merge_iters=2):
    assert len(variant_reads) > 0
    variant_seq = get_variant_nucleotides(variant_reads)
    assembly_counts = recursive_assembly(
        variant_reads,
        min_overlap_size=min_overlap_size,
        n_merge_iters=n_merge_iters)
    return [
        ((prefix, variant_seq, suffix), len(supporting_read_names))
        for ((prefix, suffix), supporting_read_names)
        in sorted(assembly_counts.items(), key=final_sort_key)
    ]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import (absolute_import,)

import os
import logging
import re

from setuptools import setup

readme_dir = os.path.dirname(__file__)
readme_filename = os.path.join(readme_dir, 'README.md')

try:
    with open(readme_filename, 'r') as f:
        readme = f.read()
except:
    logging.warn("Failed to load %s" % readme_filename)
    readme = ""

try:
    import pypandoc
    readme = pypandoc.convert(readme, to='rst', format='md')
except:
    logging.warn("Conversion of long_description from MD to RST failed")
    pass


with open('isovar/__init__.py', 'r') as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        f.read(),
        re.MULTILINE).group(1)

if not version:
    raise RuntimeError("Cannot find version information")

if __name__ == '__main__':
    setup(
        name='isovar',
        version=version,
        description="Assemble transcript sequences fragments near variants",
        author="Alex Rubinsteyn and Arman Aksoy",
        author_email="alex {dot} rubinsteyn {at} mssm {dot} edu",
        url="https://github.com/hammerlab/isovar",
        license="http://www.apache.org/licenses/LICENSE-2.0.html",
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Operating System :: OS Independent',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
        ],
        install_requires=[
            'six',
            'pysam >= 0.9.0',
            'pandas',
            'scikit-bio>=0.4.2',
            'varcode>=0.4.6'
        ],

        long_description=readme,
        packages=['isovar'],
        scripts=[
            "script/isovar-protein-sequences.py",
            "script/isovar-translations.py",
            "script/isovar-reference-contexts.py",
            "script/isovar-variant-reads.py",
            "script/isovar-variant-sequences.py",
        ],
    )



#!/usr/bin/env python

# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

import argparse

import varcode
from pysam import AlignmentFile

from isovar.args import add_somatic_vcf_args
from isovar.variant_sequence import variant_sequences_dataframe
from isovar.default_parameters import (
    VARIANT_CDNA_SEQUENCE_LENGTH,
    MIN_READS_SUPPORTING_VARIANT_CDNA_SEQUENCE
)

parser = argparse.ArgumentParser()
parser = add_somatic_vcf_args(parser)

parser.add_argument(
    "--bam",
    required=True,
    help="Path to BAM or SAM file containing RNAseq reads")

parser.add_argument(
    "--min-reads",
    type=int,
    default=MIN_READS_SUPPORTING_VARIANT_CDNA_SEQUENCE,
    help="Minimum number of reads supporting a variant sequence")

parser.add_argument(
    "--sequence-length",
    default=VARIANT_CDNA_SEQUENCE_LENGTH,
    type=int)

parser.add_argument(
    "--output",
    default="isovar-variant-sequences-results.csv",
    help="Name of CSV file which contains predicted sequences")

if __name__ == "__main__":
    args = parser.parse_args()

    print(args)

    variants = varcode.load_vcf(
        args.vcf,
        genome=args.genome)

    samfile = AlignmentFile(args.bam)

    df = variant_sequences_dataframe(
        variants=variants,
        samfile=samfile,
        sequence_length=args.sequence_length,
        min_reads=args.min_reads)

    print(df)
    df.to_csv(args.output)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import
import logging

from varcode import EffectCollection

def predicted_coding_effects_with_mutant_sequence(
        variant,
        transcript_id_whitelist=None):
    """
    For a given variant, return the set of predicted mutation effects
    on transcripts where this variant results in a predictable non-synonymous
    change to the protein sequence.

    Parameters
    ----------
    variant : varcode.Variant

    transcript_id_whitelist : set
        Filter effect predictions to only include these transcripts

    Returns a varcode.EffectCollection object
    """

    effects = []
    for transcript in variant.transcripts:
        if not transcript.complete:
            logging.info(
                "Skipping transcript %s for variant %s because it's incomplete" % (
                    transcript,
                    variant))
            continue

        if transcript_id_whitelist and transcript.id not in transcript_id_whitelist:
            logging.info(
                "Skipping transcript %s for variant %s because it's not one of %d allowed" % (
                    transcript,
                    variant,
                    len(transcript_id_whitelist)))
            continue
        effects.append(variant.effect_on_transcript(transcript))

    effects = EffectCollection(effects)

    n_total_effects = len(effects)
    logging.info("Predicted %d effects for variant %s" % (
        n_total_effects,
        variant))

    nonsynonymous_coding_effects = effects.drop_silent_and_noncoding()
    logging.info(
        "Keeping %d/%d non-synonymous coding effects for %s" % (
            len(nonsynonymous_coding_effects),
            n_total_effects,
            variant))

    usable_effects = [
        effect
        for effect in nonsynonymous_coding_effects
        if effect.mutant_protein_sequence is not None
    ]
    logging.info(
        "Keeping %d/%d effects with predictable AA sequences for %s" % (
            len(usable_effects),
            len(nonsynonymous_coding_effects),
            variant))
    return usable_effects

def reference_transcripts_for_variant(variant, transcript_id_whitelist=None):
    """
    For a given variant, find all the transcripts which overlap the
    variant and for which it has a predictable effect on the amino acid
    sequence of the protein.
    """
    predicted_effects = predicted_coding_effects_with_mutant_sequence(
        variant=variant,
        transcript_id_whitelist=transcript_id_whitelist)
    return [effect.transcript for effect in predicted_effects]



from isovar.variant_read import gather_reads_for_single_variant
from varcode import Variant
from testing_helpers import load_bam
from nose.tools import eq_

def test_somatic_variant_with_2_supporting_rna_reads():
    variant = Variant("14", 105849746, "G", "A")
    base_dir = "data/somatic-variant-with-2-supporting-rna-reads/"
    normal_reads = load_bam(base_dir + "normal.14.105849746.G.A.no-alt.sorted.bam")
    tumor_reads = load_bam(base_dir + "tumor.14.105849746.G.A.many-alt.sorted.bam")
    rna_reads = load_bam(base_dir + "rna.14.105849746.G.A.2-alt.sorted.bam")

    normal_sample_variant_reads = gather_reads_for_single_variant(
        variant=variant,
        samfile=normal_reads)
    eq_(len(normal_sample_variant_reads), 0)
    print(normal_sample_variant_reads)

    tumor_sample_variant_reads = gather_reads_for_single_variant(
        variant=variant,
        samfile=tumor_reads)
    print(tumor_sample_variant_reads)
    eq_(len(tumor_sample_variant_reads), 8)

    rna_sample_variant_reads = gather_reads_for_single_variant(
        variant=variant,
        samfile=rna_reads)
    print(rna_sample_variant_reads)
    eq_(len(rna_sample_variant_reads), 2)
    # Arun went through the hassle of pulling out the exact read names
    # in IGV
    expected_variant_rna_read_names = {
        "K00193:50:H5NKVBBXX:5:2202:6421:24964",
        "K00193:50:H5NKVBBXX:5:2119:30908:1138",
    }
    for variant_read in rna_sample_variant_reads:
        assert variant_read.name in expected_variant_rna_read_names

if __name__ == "__main__":
    test_somatic_variant_with_2_supporting_rna_reads()
